# coding: utf-8

"""
Preview Feature Test API

Test spec for preview feature filtering  # noqa: E501

The version of the OpenAPI document: 1.0.0
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations
import pprint
import re  # noqa: F401
import json


from typing import Union, Optional


from datetime import datetime

from datetime import datetime

from pydantic import BaseModel, ConfigDict, StrictStr

from typing import Any, ClassVar, Dict, List, Optional


class Resource(BaseModel):
    """A model object representing the Resource resource.

    Constructs an object of type Resource with the provided properties.

    Parameters
    __________
    name : str, optional


    status : str, optional


    created_at : datetime, optional


    """

    name: Optional[StrictStr] = None

    status: Optional[StrictStr] = None

    created_at: Optional[datetime] = None

    __properties = ["name", "status", "created_at"]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Resource:
        """Create an instance of Resource from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias"""

        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = dict(self._iter(to_dict=True, by_alias=True, exclude=exclude_properties, exclude_none=True))

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> Resource:
        """Create an instance of Resource from a dict."""

        if obj is None:
            return None

        if type(obj) is not dict:
            return Resource.parse_obj(obj)

        _obj = Resource.parse_obj(
            {
                "name": obj.get("name"),
                "status": obj.get("status"),
                "created_at": obj.get("created_at"),
            }
        )

        return _obj


class ResourceModel:
    def __init__(
        self,  # optional properties
        name: Optional[str] = None,
        status: Optional[str] = None,
        created_at: Optional[datetime] = None,
    ):
        """A model object representing the Resource resource.

        Constructs an object of type Resource with the provided properties.

        Parameters
        __________
        name : str, optional

        status : str, optional

        created_at : datetime, optional

        """

        self.name = name
        self.status = status
        self.created_at = created_at

    __properties = ["name", "status", "created_at"]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return Resource(
            name=self.name,
            status=self.status,
            created_at=self.created_at,
        )

    @classmethod
    def _from_model(cls, model) -> ResourceModel:
        return ResourceModel(
            name=model.name,
            status=model.status,
            created_at=model.created_at,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a Resource.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the Resource object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> ResourceModel:
        """Create an instance of Resource from a dict.

        This method constructs a Resource object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        Resource
            A Resource object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(Resource.from_dict(obj))


Resource._model_class = ResourceModel



"""
Snowflake Sequence API.

The Snowflake Sequence API is a REST API that you can use to access, update, and perform certain actions on Sequence resource in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from typing import TYPE_CHECKING, Any, Iterator, Optional, TypeVar, Union

from snowflake.core._common import CreateMode, SchemaObjectCollectionParent, SchemaObjectReferenceMixin
from snowflake.core._generated.api_client import StoredProcApiClient
from snowflake.core._internal.telemetry import api_telemetry
from snowflake.core._operation import PollingOperation, PollingOperations
from snowflake.core._utils import get_function_name_with_args, replace_function_name_in_name_with_args
from snowflake.core.sequence._generated.models.sequence import Sequence

from .sequence_api import SequenceApi


if TYPE_CHECKING:
    from snowflake.core.schema import SchemaResource


T = TypeVar("T")


class SequenceCollectionBase(SchemaObjectCollectionParent["SequenceResource"]):
    _identifier_requires_args: bool = False

    def __init__(self, schema: "SchemaResource", resource_class: type[T]) -> None:
        super().__init__(schema, resource_class)
        self._api = SequenceApi(
            root=self.root, resource_class=self._ref_class, sproc_client=StoredProcApiClient(root=self.root)
        )

    @api_telemetry
    def create(self, sequence: Sequence, mode: Optional[Union[CreateMode, str]] = None) -> "SequenceResource":  # noqa: F821
        """Create a sequence.

        Parameters
        __________
        sequence: Sequence
             (required)
        mode: Union[CreateMode, str]
             Parameter allowing support for different modes of resource creation. Possible values include: - `errorIfExists`: Throws an error if you try to create a resource that already exists. - `orReplace`: Automatically replaces the existing resource with the current one. - `ifNotExists`: Creates a new resource when an alter is requested for a non-existent resource.
        """
        if isinstance(mode, str):
            mode = CreateMode[mode].value

        self._api.create_sequence(
            self.database.name,
            self.schema.name,
            sequence=sequence,
            create_mode=mode,
            async_req=False,
        )
        if self._identifier_requires_args:
            return self._ref_class(get_function_name_with_args(sequence), self)
        return self._ref_class(sequence.name, self)

    @api_telemetry
    def create_async(
        self, sequence: Sequence, mode: Optional[Union[CreateMode, str]] = None
    ) -> PollingOperation["SequenceResource"]:  # noqa: F821
        """An asynchronous version of :func:`create`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """
        if isinstance(mode, str):
            mode = CreateMode[mode].value

        future = self._api.create_sequence(
            self.database.name,
            self.schema.name,
            sequence=sequence,
            create_mode=mode,
            async_req=True,
        )
        if self._identifier_requires_args:
            return PollingOperation(future, lambda _: self._ref_class(get_function_name_with_args(sequence), self))
        return PollingOperation(
            future, lambda _: self._ref_class(sequence if isinstance(sequence, str) else sequence.name, self)
        )

    @api_telemetry
    def iter(self, *, like: Optional[str] = None) -> Iterator[Sequence]:
        """List sequences.

        Parameters
        __________
        like: str
             Parameter to filter the command output by resource name. Uses case-insensitive pattern matching, with support for SQL wildcard characters.
        """
        resources = self._api.list_sequences(
            self.database.name,
            self.schema.name,
            like=like,
            async_req=False,
        )
        return iter(resources)

    @api_telemetry
    def iter_async(self, *, like: Optional[str] = None) -> PollingOperation[Iterator[Sequence]]:
        """An asynchronous version of :func:`iter`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """
        future = self._api.list_sequences(
            self.database.name,
            self.schema.name,
            like=like,
            async_req=True,
        )

        return PollingOperations.iterator(future)


class SequenceResourceBase(SchemaObjectReferenceMixin["SequenceCollection"]):
    _collection_class: SequenceCollectionBase
    _identifier_requires_args: bool = False
    _plural_name: str

    def __init__(self, name: str, collection_class: SequenceCollectionBase) -> None:
        self.collection = collection_class
        if self._identifier_requires_args:
            self.name_with_args = name
        else:
            self.name = name

    @api_telemetry
    def clone(
        self,
        sequence: Sequence,
        create_mode: Optional[Union[CreateMode, str]] = None,
        target_database: Optional[str] = None,
        target_schema: Optional[str] = None,
    ) -> None:
        """Create a new sequence by cloning from the specified resource.

        Parameters
        __________
        sequence: Sequence
             (required)
        create_mode: Union[CreateMode, str]
             Parameter allowing support for different modes of resource creation. Possible values include: - `errorIfExists`: Throws an error if you try to create a resource that already exists. - `orReplace`: Automatically replaces the existing resource with the current one. - `ifNotExists`: Creates a new resource when an alter is requested for a non-existent resource.
        target_database: str
             Database of the target resource. Defaults to the source's database
        target_schema: str
             Schema of the target resource. Defaults to the source's schema
        """
        self.collection._api.clone_sequence(
            self.database.name,
            self.schema.name,
            self._identifier,
            sequence=sequence,
            create_mode=create_mode,
            target_database=target_database,
            target_schema=target_schema,
            async_req=False,
        )

    @api_telemetry
    def clone_async(
        self,
        sequence: Sequence,
        create_mode: Optional[Union[CreateMode, str]] = None,
        target_database: Optional[str] = None,
        target_schema: Optional[str] = None,
    ) -> PollingOperation[None]:
        """An asynchronous version of :func:`clone`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """
        future = self.collection._api.clone_sequence(
            self.database.name,
            self.schema.name,
            self._identifier,
            sequence=sequence,
            create_mode=create_mode,
            target_database=target_database,
            target_schema=target_schema,
            async_req=True,
        )
        return PollingOperations.empty(future)

    @api_telemetry
    def fetch(
        self,
    ) -> Sequence:
        """Fetch a sequence.

        Parameters
        __________
        """
        result = self.collection._api.fetch_sequence(
            self.database.name,
            self.schema.name,
            self._identifier,
            async_req=False,
        )
        return result

    @api_telemetry
    def fetch_async(
        self,
    ) -> PollingOperation[Sequence]:
        """An asynchronous version of :func:`fetch`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """
        future = self.collection._api.fetch_sequence(
            self.database.name,
            self.schema.name,
            self._identifier,
            async_req=True,
        )

        return PollingOperations.identity(future)

    @api_telemetry
    def rename(
        self,
        target_name: str,
        if_exists: Optional[bool] = None,
        target_database: Optional[str] = None,
        target_schema: Optional[str] = None,
    ) -> None:
        """Rename a sequence with a new identifier.

        Parameters
        __________
        target_name: str
             Name of the target resource. (required)
        if_exists: bool
             Parameter that specifies how to handle the request for a resource that does not exist: - `true`: The endpoint does not throw an error if the resource does not exist. It returns a 200 success response, but does not take any action on the resource. - `false`: The endpoint throws an error if the resource doesn't exist.
        target_database: str
             Database of the target resource. Defaults to the source's database
        target_schema: str
             Schema of the target resource. Defaults to the source's schema
        """
        self.collection._api.rename_sequence(
            self.database.name,
            self.schema.name,
            self._identifier,
            target_name=target_name,
            if_exists=if_exists,
            target_database=target_database,
            target_schema=target_schema,
            async_req=False,
        )

        self._rename_finalizer(target_name, target_schema, target_database)

    @api_telemetry
    def rename_async(
        self,
        target_name: str,
        if_exists: Optional[bool] = None,
        target_database: Optional[str] = None,
        target_schema: Optional[str] = None,
    ) -> PollingOperation[None]:
        """An asynchronous version of :func:`rename`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """
        future = self.collection._api.rename_sequence(
            self.database.name,
            self.schema.name,
            self._identifier,
            target_name=target_name,
            if_exists=if_exists,
            target_database=target_database,
            target_schema=target_schema,
            async_req=True,
        )

        def finalize(_: Any) -> None:
            self._rename_finalizer(
                target_name=target_name, target_schema=target_schema, target_database=target_database
            )

        return PollingOperation(future, finalize)

    def _rename_finalizer(self, target_name: str, target_schema: Optional[str], target_database: Optional[str]) -> None:
        """Finalizer for rename operation."""
        self._set_new_name(target_name)
        if target_database is None:
            target_database = self.database.name
        if target_schema is None:
            target_schema = self.schema.name

        # Schema level resource
        if target_database != self.database.name or target_schema != self.schema.name:
            self.collection = getattr(self.root.databases[target_database].schemas[target_schema], self._plural_name)

    @api_telemetry
    def drop(
        self,
        if_exists: Optional[bool] = None,
    ) -> None:
        """Delete a sequence.

        Parameters
        __________
        if_exists: bool
             Parameter that specifies how to handle the request for a resource that does not exist: - `true`: The endpoint does not throw an error if the resource does not exist. It returns a 200 success response, but does not take any action on the resource. - `false`: The endpoint throws an error if the resource doesn't exist.
        """
        self.collection._api.delete_sequence(
            self.database.name,
            self.schema.name,
            self._identifier,
            if_exists=if_exists,
            async_req=False,
        )

    @api_telemetry
    def drop_async(
        self,
        if_exists: Optional[bool] = None,
    ) -> PollingOperation[None]:
        """An asynchronous version of :func:`drop`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """
        future = self.collection._api.delete_sequence(
            self.database.name,
            self.schema.name,
            self._identifier,
            if_exists=if_exists,
            async_req=True,
        )
        return PollingOperations.empty(future)

    @property
    def _identifier(self) -> str:
        """Get the identifier for this resource."""
        if self._identifier_requires_args:
            return self.name_with_args
        return self.name

    def _set_new_name(self, target_name: str) -> None:
        """Set a new name for the resource."""
        if self._identifier_requires_args:
            old_name = self.name_with_args
            self.name_with_args = replace_function_name_in_name_with_args(self.name_with_args, target_name)
        else:
            old_name = self.name
            self.name = target_name

        # Update the collection reference to point to the new name
        self.collection.update_reference(old_name=old_name, new_name=self._identifier, resource=self)



from typing import TYPE_CHECKING

from ._generated.api.secret_api_base import SecretCollectionBase, SecretResourceBase


if TYPE_CHECKING:
    from snowflake.core.schema import SchemaResource


class SecretCollection(SecretCollectionBase):
    """Represents the collection operations on the Snowflake secret resource.

    With this collection, you can create, iterate through, and search for secrets that you have access to
    in the current context.
    """

    def __init__(self, schema: "SchemaResource"):
        super().__init__(schema, SecretResource)


class SecretResource(SecretResourceBase):
    """Represents a reference to a Snowflake secret.

    With this secret reference, you can fetch information about a secret, as well as perform
    certain actions on it.
    """

    _plural_name = "secrets"



import pytest

from snowflake.core.account import AccountCollection


@pytest.fixture(scope="session")
def temp_customer_account_accounts(temp_customer_account_root) -> AccountCollection:
    return temp_customer_account_root.accounts



import time

from contextlib import suppress

import pytest

from snowflake.core import Clone, PointOfTimeOffset
from snowflake.core.dynamic_table import (
    DownstreamLag,
    DynamicTable,
    DynamicTableClone,
    DynamicTableColumn,
    UserDefinedLag,
)
from snowflake.core.exceptions import NotFoundError
from snowflake.core.schema import Schema
from tests.utils import random_string


def test_create_from_model_create_mode(dynamic_tables, db_parameters, table_handle):
    table_name = '"' + random_string(10, "test_table_ INTEGRATION_") + '"'
    created_handle = dynamic_tables.create(
        DynamicTable(
            name=table_name,
            columns=[
                DynamicTableColumn(name="c1", comment="column comment"),
                DynamicTableColumn(name='"cc2"', datatype="varchar", comment="column comment"),
            ],
            warehouse=db_parameters["warehouse"],
            refresh_mode="FULL",
            initialize="ON_CREATE",
            comment="test comment",
            target_lag=UserDefinedLag(seconds=60),
            query=f"SELECT * FROM {table_handle.name}",
        ),
        mode="errorifexists",
    )
    try:
        assert created_handle.name == table_name
        assert created_handle.fetch().target_lag == UserDefinedLag(seconds=60)
        assert created_handle.fetch().comment == "test comment"
        assert created_handle.fetch().refresh_mode == "FULL"
        assert created_handle.fetch().initialize == "ON_CREATE"
        assert created_handle.fetch().query == f"SELECT * FROM {table_handle.name}"
        assert created_handle.fetch().warehouse.replace('"', "") == db_parameters["warehouse"].upper().replace('"', "")
        assert created_handle.fetch().columns[0].name.lower() == "c1"
        assert created_handle.fetch().columns[0].comment == "column comment"
        assert created_handle.fetch().columns[1].comment == "column comment"
        assert "VARCHAR" in created_handle.fetch().columns[1].datatype
        assert created_handle.fetch().columns[1].name.lower() == "cc2"
    finally:
        with suppress(Exception):
            created_handle.drop()


@pytest.mark.min_sf_ver("8.27.0")
def test_create_clone(dynamic_tables, dynamic_table_handle, session):
    table_name = random_string(10, "test_table_")
    created_handle = dynamic_tables.create(
        table_name, clone_table=f"{dynamic_table_handle.name}", copy_grants=True, mode="errorifexists"
    )
    try:
        assert created_handle.name == table_name
        assert created_handle.fetch().target_lag == DownstreamLag()
    finally:
        with suppress(Exception):
            created_handle.drop()

    time.sleep(1)
    table_name = random_string(10, "test_table_")
    created_handle = dynamic_tables.create(
        DynamicTableClone(name=table_name, target_lag=UserDefinedLag(seconds=120)),
        clone_table=Clone(
            source=f"{dynamic_table_handle.name}", point_of_time=PointOfTimeOffset(reference="before", when="-1")
        ),
        copy_grants=True,
        mode="errorifexists",
    )
    try:
        assert created_handle.name == table_name
        assert created_handle.fetch().target_lag == UserDefinedLag(seconds=120)
    finally:
        with suppress(Exception):
            created_handle.drop()

    table_name = random_string(10, "test_table_")
    with pytest.raises(NotFoundError):
        dynamic_tables.create(table_name, clone_table="non_existant_name", copy_grants=True, mode="errorifexists")


def test_create_clone_across_schemas(dynamic_table_handle, temp_schema):
    # clone <current_schema>.dynamic_table to temp_schema.table_name
    table_name = random_string(10, "test_clone_table_across_schema_")
    created_handle = temp_schema.dynamic_tables.create(
        DynamicTableClone(name=table_name, target_lag=UserDefinedLag(seconds=120)),
        clone_table=Clone(source=f"{dynamic_table_handle.schema.name}.{dynamic_table_handle.name}"),
        mode="errorifexists",
    )

    try:
        assert created_handle.name == table_name
        assert created_handle.fetch().target_lag == UserDefinedLag(seconds=120)
        assert created_handle.fetch().schema_name.upper() == temp_schema.name.upper()
    finally:
        with suppress(Exception):
            created_handle.drop()


def test_create_clone_across_database(dynamic_table_handle, temp_db):
    # clone <current_db>.<current_schema>.dynamic_table to temp_db.created_schema.table_name
    schema_name = random_string(10, "test_create_clone_across_database_schema_")
    created_schema = temp_db.schemas.create(Schema(name=schema_name))
    table_name = random_string(10, "test_table_clone_across_database_table_")

    try:
        created_handle = created_schema.dynamic_tables.create(
            DynamicTableClone(name=table_name, target_lag=UserDefinedLag(seconds=120)),
            clone_table=Clone(
                source=f"{dynamic_table_handle.database.name}.{dynamic_table_handle.schema.name}.{dynamic_table_handle.name}"
            ),
            mode="errorifexists",
        )

        assert created_handle.name == table_name
        assert created_handle.fetch().target_lag == UserDefinedLag(seconds=120)
        assert created_handle.fetch().schema_name.upper() == created_schema.name.upper()
        assert created_handle.fetch().database_name.upper() == temp_db.name.upper()
    finally:
        with suppress(Exception):
            created_schema.drop()



# coding: utf-8

# flake8: noqa

"""
Snowflake External Volume API.

The Snowflake External Volume API is a REST API that you can use to access, update, and perform certain actions on External Volume resource in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import absolute_import

__version__ = "1.0.0"

# import apis into sdk package
from snowflake.core.external_volume._generated.api.external_volume_api import ExternalVolumeApi

# import models into sdk package
from snowflake.core.external_volume._generated.models.encryption import Encryption
from snowflake.core.external_volume._generated.models.error_response import ErrorResponse
from snowflake.core.external_volume._generated.models.external_volume import ExternalVolume
from snowflake.core.external_volume._generated.models.storage_location import StorageLocation
from snowflake.core.external_volume._generated.models.storage_location_azure import StorageLocationAzure
from snowflake.core.external_volume._generated.models.storage_location_gcs import StorageLocationGcs
from snowflake.core.external_volume._generated.models.storage_location_s3 import StorageLocationS3
from snowflake.core.external_volume._generated.models.storage_location_s3_gov import StorageLocationS3Gov
from snowflake.core.external_volume._generated.models.success_accepted_response import SuccessAcceptedResponse
from snowflake.core.external_volume._generated.models.success_response import SuccessResponse



import pytest

from snowflake.core.exceptions import NotFoundError
from snowflake.core.streamlit import Streamlit
from tests.integ.utils import random_string


pytestmark = [pytest.mark.min_sf_ver("9.38.0")]


def test_rename(streamlits, streamlit_stage_with_file, warehouse, streamlit_main_file):
    original_name = random_string(10, "test_original_streamlit_")
    new_name = random_string(10, "test_renamed_streamlit_")

    st = Streamlit(
        name=original_name,
        query_warehouse=warehouse.name,
        source_location=f"@{streamlit_stage_with_file.name}",
        main_file=streamlit_main_file,
    )

    streamlit_handle = streamlits.create(st)
    try:
        fetched_streamlit = streamlit_handle.fetch()
        assert fetched_streamlit.name.upper() == original_name.upper()

        streamlit_handle.rename(new_name)

        fetched_after_rename = streamlits[new_name].fetch()
        assert fetched_after_rename.name.upper() == new_name.upper()
        assert fetched_after_rename.database_name.upper() == streamlits.database.name.upper()
        assert fetched_after_rename.schema_name.upper() == streamlits.schema.name.upper()

        with pytest.raises(NotFoundError):
            streamlits[original_name].fetch()
    finally:
        streamlits[new_name].drop(if_exists=True)


def test_rename_nonexistent_streamlit(streamlits):
    """Test renaming a non-existent streamlit raises NotFoundError."""
    nonexistent_name = random_string(10, "test_nonexistent_streamlit_")
    new_name = random_string(10, "test_new_name_streamlit_")

    with pytest.raises(NotFoundError):
        streamlits[nonexistent_name].rename(new_name)

    streamlits[nonexistent_name].rename(new_name, if_exists=True)


def test_rename_streamlit_cross_schema(
    streamlits, streamlit_stage_with_file, warehouse, streamlit_main_file, temp_schema
):
    """Test renaming streamlit across schemas."""
    original_name = random_string(10, "test_cross_schema_streamlit_")
    new_name = random_string(10, "test_renamed_cross_schema_streamlit_")

    st = Streamlit(
        name=original_name,
        query_warehouse=warehouse.name,
        source_location=f"@{streamlit_stage_with_file.database.name}.{streamlit_stage_with_file.schema.name}.{streamlit_stage_with_file.name}",
        main_file=streamlit_main_file,
    )

    streamlit_handle = streamlits.create(st)
    original_db_name = streamlit_handle.database.name
    try:
        streamlit_handle.rename(
            new_name,
            target_schema=temp_schema.name,
            target_database=original_db_name,
        )

        fetched_streamlit = streamlit_handle.fetch()
        assert fetched_streamlit.name.upper() == new_name.upper()
        assert fetched_streamlit.schema_name.upper() == temp_schema.name.upper()
        assert fetched_streamlit.database_name.upper() == original_db_name.upper()

        with pytest.raises(NotFoundError):
            streamlits[original_name].fetch()
    finally:
        streamlit_handle.drop(if_exists=True)



from typing import Any, Optional

from snowflake.core.database_role._generated.models.grant import Grant as GrantModel
from snowflake.core.grant._grantee import Grantee
from snowflake.core.grant._privileges import Privileges
from snowflake.core.grant._securables import Securable


class Grant:
    """
    Represents Snowflake Grant Operation.

    Parameters
    __________
    grantee: Grantee
        The role that the ``Grant`` is being applied to.
    securable: Securable
        The ``Securable`` object.
    privileges: Privileges
        The ``Privileges`` being granted.
    grant_option: bool, optional
        If ``True``, grantee can pass this privilege down.
        Default is ``None``, which is equivalent to ``False``.

    Examples
    ________
    Granting privileges to a test role:

    >>> from snowflake.core.grant import Grantees, Privileges, Securables
    >>> Grant(grantee=Grantees.role(
    ...     name="test_role",
    ...     securable=Securables.current_account,
    ...     privileges=[Privileges.create_database],
    ... )
    """

    def __init__(
        self,
        grantee: Grantee,
        securable: Securable,
        privileges: Optional[list[Privileges]] = None,
        grant_option: bool = False,
    ):
        self._grantee = grantee
        self._securable = securable
        self._privileges = privileges
        self._grant_option = grant_option

    @property
    def grantee(self) -> Grantee:
        return self._grantee

    @property
    def securable(self) -> Securable:
        return self._securable

    @property
    def privileges(self) -> Optional[list[Privileges]]:
        return self._privileges

    @property
    def grant_option(self) -> bool:
        return self._grant_option

    def to_dict(self, hide_readonly_properties: bool = False) -> dict[str, Any]:
        privileges = [privilege.value for privilege in self.privileges] if self.privileges else None
        request_body = GrantModel(
            privileges=privileges, grant_option=self.grant_option, securable_type=self.securable.securable_type
        )
        return request_body.to_dict(hide_readonly_properties=hide_readonly_properties)



import pytest

from snowflake.core import Clone
from snowflake.core.schema import Schema
from snowflake.core.stream import Stream, StreamSourceTable
from tests.utils import random_string

from ...utils import ensure_snowflake_version


@pytest.fixture
def temp_stream(streams, src_temp_table, snowflake_version):
    ensure_snowflake_version(snowflake_version, "8.37.0")

    stream_name = random_string(10, "test_stream_")
    stream_handle = streams.create(
        Stream(name=stream_name, stream_source=StreamSourceTable(name=src_temp_table.name), comment="ThIs iS a ComMeNT")
    )

    try:
        yield stream_handle
    finally:
        streams[stream_name].drop()


def test_clone(streams, temp_stream):
    stream_name = random_string(10, "test_stream_clone_")

    cloned_stream = streams.create(stream_name, clone_stream=temp_stream.name)

    try:
        cloned_handle = cloned_stream.fetch()
        assert isinstance(temp_stream.fetch().stream_source, type(cloned_handle.stream_source))
        assert temp_stream.fetch().stream_source.name == cloned_handle.stream_source.name
        assert cloned_handle.comment == "ThIs iS a ComMeNT"
    finally:
        streams[stream_name].drop()


def test_clone_with_point_of_time(streams, temp_stream):
    stream_name = random_string(10, "test_stream_clone_pot_")
    cloned_stream = streams.create(stream_name, clone_stream=Clone(source=temp_stream.name))

    try:
        cloned_handle = cloned_stream.fetch()
        assert isinstance(temp_stream.fetch().stream_source, type(cloned_handle.stream_source))
        assert temp_stream.fetch().stream_source.name == cloned_handle.stream_source.name
        assert cloned_handle.comment == "ThIs iS a ComMeNT"
    finally:
        streams[stream_name].drop()


def test_clone_across_schema(temp_stream, temp_schema):
    stream_name = random_string(10, "test_clone_stream_across_schema_")

    created_handle = temp_schema.streams.create(
        stream_name, clone_stream=f"{temp_stream.schema.name}.{temp_stream.name}"
    )

    try:
        assert isinstance(temp_stream.fetch().stream_source, type(created_handle.fetch().stream_source))
        assert temp_stream.fetch().stream_source.name == created_handle.fetch().stream_source.name
        assert created_handle.fetch().database_name.upper() == temp_schema.database.name.upper()
        assert created_handle.fetch().schema_name.upper() == temp_schema.name.upper()
        assert created_handle.fetch().comment == "ThIs iS a ComMeNT"
    finally:
        temp_schema.streams[stream_name].drop()


def test_clone_across_database(temp_stream, temp_db):
    schema_name = random_string(10, "test_create_clone_across_schema_")
    created_schema = temp_db.schemas.create(Schema(name=schema_name))
    stream_name = random_string(10, "test_stream_clone_across_database_")

    try:
        created_handle = created_schema.streams.create(
            stream_name, clone_stream=f"{temp_stream.database.name}.{temp_stream.schema.name}.{temp_stream.name}"
        )

        assert isinstance(temp_stream.fetch().stream_source, type(created_handle.fetch().stream_source))
        assert temp_stream.fetch().stream_source.name == created_handle.fetch().stream_source.name
        assert created_handle.fetch().database_name.upper() == created_schema.database.name.upper()
        assert created_handle.fetch().schema_name.upper() == created_schema.name.upper()
        assert created_handle.fetch().comment == "ThIs iS a ComMeNT"
    finally:
        created_schema.drop()



import copy

from collections.abc import Generator

import pytest as pytest

from snowflake.core.table import Table, TableColumn, TableResource
from snowflake.core.view import View, ViewColumn
from tests.integ.utils import random_string


test_table_template = Table(
    name="<to be set>",
    columns=[
        TableColumn(
            name="c1",
            datatype="int",
            nullable=False,
            autoincrement=True,
            autoincrement_start=0,
            autoincrement_increment=1,
        ),
        TableColumn(name="c2", datatype="string"),
        TableColumn(name="c3", datatype="string", collate="FR"),
    ],
)

test_view_template = View(
    name="<to be set>",
    columns=[ViewColumn(name="c1"), ViewColumn(name="c2"), ViewColumn(name="c3")],
    query="<to be set>",
)


@pytest.fixture(scope="module")
def temp_table(tables) -> Generator[TableResource, None, None]:
    table_name = f"test_table_for_view_{random_string(10)}"
    test_table = copy.deepcopy(test_table_template)
    test_table.name = table_name
    test_table_handle = tables.create(test_table)
    try:
        yield test_table_handle
    finally:
        test_table_handle.drop()



"""Manages Snowflake database roles.

Example:
    >>> database_role_name = "test_database_role"
    >>> database_role = DatabaseRole(name=database_role_name, comment="test_comment")
    >>> created_database_role = database_roles.create(database_role)
    >>> database_roles[database_role_name].drop()
"""

from ._database_role import (
    ContainingScope,
    DatabaseRole,
    DatabaseRoleCollection,
    DatabaseRoleResource,
    Grant,
    Securable,
)


__all__ = ["DatabaseRole", "DatabaseRoleCollection", "DatabaseRoleResource", "ContainingScope", "Securable", "Grant"]



import pytest

from snowflake.core.exceptions import NotFoundError


pytestmark = [pytest.mark.min_sf_ver("8.35.0")]


def test_drop_pipe(pipes, template_pipe):
    try:
        pipe_resource = pipes.create(template_pipe)
        pipe_names = [pipe.name for pipe in pipes.iter(like="test_create_pipe_%")]
        assert template_pipe.name.upper() in pipe_names

        pipe_resource.drop()
        pipe_names = [pipe.name for pipe in pipes.iter(like="test_create_pipe_%")]
        assert len(pipe_names) == 0

        # drop the already dropped pipe
        with pytest.raises(NotFoundError):
            pipe_resource.drop()

        # drop the already dropped pipe with if_exists
        pipe_resource.drop(if_exists=True)
    finally:
        pipe_resource.drop(if_exists=True)



"""
Snowflake Catalog Integration API.

The Snowflake Catalog Integration API is a REST API that you can use to access, update, and perform certain actions on Catalog Integration resource in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, List

from pydantic import ConfigDict, StrictStr

from snowflake.core.catalog_integration._generated.models.rest_authentication import (
    RestAuthentication,
)


class OAuth(RestAuthentication):
    """A model object representing the OAuth resource.

    Constructs an object of type OAuth with the provided properties.

    Parameters
    __________
    oauth_client_id : str
        The client ID of the OAuth2 credential associated with the Polaris service connection.
    oauth_client_secret : str
        The secret for the OAuth2 credential associated with the Polaris service connection.
    oauth_allowed_scopes : list[str]
        The scope of the OAuth token. Only one scope is included in the Iceberg REST API specification, but catalogs can support more than one scope in their implementation.
    """

    oauth_client_id: StrictStr

    oauth_client_secret: StrictStr

    oauth_allowed_scopes: List[StrictStr]

    __properties = ["type"]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> OAuth:
        """Create an instance of OAuth from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        _dict["type"] = RestAuthentication.get_child_model_discriminator_value("OAuth")

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> OAuth:
        """Create an instance of OAuth from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return OAuth.model_validate(obj)

        _obj = OAuth.model_validate(
            {
                "oauth_client_id": obj.get("oauth_client_id"),
                "oauth_client_secret": obj.get("oauth_client_secret"),
                "oauth_allowed_scopes": obj.get("oauth_allowed_scopes"),
            }
        )

        return _obj


class OAuthModel(RestAuthentication):
    def __init__(
        self,
        oauth_client_id: str,
        oauth_client_secret: str,
        oauth_allowed_scopes: list[str],
        # optional properties
    ):
        """A model object representing the OAuth resource.

        Constructs an object of type OAuth with the provided properties.

        Parameters
        __________
        oauth_client_id : str
            The client ID of the OAuth2 credential associated with the Polaris service connection.
        oauth_client_secret : str
            The secret for the OAuth2 credential associated with the Polaris service connection.
        oauth_allowed_scopes : list[str]
            The scope of the OAuth token. Only one scope is included in the Iceberg REST API specification, but catalogs can support more than one scope in their implementation.
        """
        super().__init__()
        self.oauth_client_id = oauth_client_id
        self.oauth_client_secret = oauth_client_secret
        self.oauth_allowed_scopes = oauth_allowed_scopes

    __properties = ["type"]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return OAuth(
            oauth_client_id=self.oauth_client_id,
            oauth_client_secret=self.oauth_client_secret,
            oauth_allowed_scopes=self.oauth_allowed_scopes,
        )

    @classmethod
    def _from_model(cls, model) -> OAuthModel:
        return OAuthModel(
            oauth_client_id=model.oauth_client_id,
            oauth_client_secret=model.oauth_client_secret,
            oauth_allowed_scopes=model.oauth_allowed_scopes,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a OAuth.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the OAuth object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> OAuthModel:
        """Create an instance of OAuth from a dict.

        This method constructs a OAuth object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        OAuth
            A OAuth object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(OAuth.from_dict(obj))


OAuth._model_class = OAuthModel



#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#

from io import BytesIO
from textwrap import dedent
from time import sleep

import pytest

from snowflake.core.exceptions import APIError
from snowflake.core.service import Service, ServiceSpecStageFile
from tests.utils import random_string


pytestmark = [pytest.mark.skip_gov]


def test_fetch(services, temp_service, database, shared_compute_pool):
    service = services[temp_service.name].fetch()
    assert service.name == temp_service.name.upper()
    assert service.compute_pool == shared_compute_pool.upper()
    assert service.auto_resume
    assert service.comment == "created by temp_service"
    assert service.database_name == database.name
    assert service.spec is not None
    assert service.current_instances is not None
    assert service.target_instances is not None
    assert service.min_instances is not None
    assert service.max_instances is not None
    assert service.owner is not None
    assert service.dns_name is not None
    assert service.created_on is not None
    assert service.updated_on is not None
    assert service.owner_role_type is not None
    assert service.is_job is not None
    assert service.spec_digest is not None
    assert service.is_upgrading is not None


@pytest.mark.flaky
def test_fetch_service_logs(services, temp_service, instance_family):
    service_name = temp_service.name
    logs = None

    for _ in range(10):
        try:
            logs = services[service_name].get_service_logs("0", "hello-world")
        except APIError:
            pass
        if logs:
            break
        sleep(5)

    if instance_family.lower() != "fake":
        message = "your installation appears to be working"
    else:
        message = "test 999"
    assert message in logs

    trimmed_logs = services[temp_service.name].get_service_logs("0", "hello-world", num_lines=10)
    assert trimmed_logs in logs
    assert len(trimmed_logs) < len(logs)


def test_fetch_service_status(services, session, imagerepo, shared_compute_pool):
    service_name = random_string(5, "test_service_")
    stage_name = random_string(5, "test_stage_")
    session.sql(f"create temp stage {stage_name};").collect()
    spec_file = "spec.yaml"
    stage_file = f"@{stage_name}"
    spec = f"{stage_file}/{spec_file}"
    session.file.put_stream(
        BytesIO(
            dedent(f"""
                spec:
                  containers:
                  - name: hello-world
                    image: {imagerepo}/hello-world:latest
                """).encode()
        ),
        spec,
    )
    test_service = Service(
        name=service_name,
        compute_pool=shared_compute_pool,
        spec=ServiceSpecStageFile(stage=stage_name, spec_file=spec_file),
        min_instances=1,
        max_instances=1,
    )

    try:
        s = services.create(test_service)
        status = s.get_service_status(timeout=10)
        assert status[0]["status"] in ["UNKNOWN", "PENDING", "READY", "DONE"]
        status = s.get_service_status()
        assert status[0]["status"] in ["UNKNOWN", "PENDING", "READY", "DONE"]
        s.suspend()
        for _ in range(10):
            status = s.get_service_status(timeout=10)
            if not status:
                break
            sleep(1)
        else:
            pytest.fail(f"Service status did not become empty after suspend. Last status: {status}")
        assert not status
        status = s.get_service_status()
        assert not status
    finally:
        services[test_service.name].drop()



import pytest

from snowflake.core.managed_account import ManagedAccountCollection


@pytest.fixture(scope="session")
def temp_customer_account_managed_accounts(temp_customer_account_root) -> ManagedAccountCollection:
    return temp_customer_account_root.managed_accounts



from typing import TYPE_CHECKING

from snowflake.core import PollingOperation
from snowflake.core._generated.api_client import StoredProcApiClient
from snowflake.core._internal.telemetry import api_telemetry
from snowflake.core.cortex.chat_service._generated.api import CortexChatApi
from snowflake.core.cortex.chat_service._generated.models import ChatRequest
from snowflake.core.rest import SSEClient


if TYPE_CHECKING:
    from snowflake.core import Root


class CortexChatService:
    """Represents the collection operations of the Snowflake Cortex Chat Service resource."""

    def __init__(self, root: "Root") -> None:
        self._api = CortexChatApi(root=root, resource_class=None, sproc_client=StoredProcApiClient(root=root))

    @api_telemetry
    def chat(self, chat_request: ChatRequest) -> SSEClient:
        """Perform chat service, similar to snowflake.cortex.Chat.

        Parameters
        __________
        chat_request: ChatRequest
            The chat request object to be sent to the chat service. Defined in ./_generated/models/chat_request.py
        """
        return SSEClient(self._api.chat_request(chat_request, async_req=False, _preload_content=False))

    @api_telemetry
    def chat_async(self, chat_request: ChatRequest) -> PollingOperation[SSEClient]:
        """An asynchronous version of :func:`chat`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        future = self._api.chat_request(chat_request, async_req=True, _preload_content=False)
        return PollingOperation(future, lambda x: SSEClient(x))



from unittest import mock

from snowflake.core import PollingOperation
from snowflake.core.tag import Tag, TagResource

from ...utils import BASE_URL, extra_params, mock_http_response


API_CLIENT_REQUEST = "snowflake.core._generated.api_client.ApiClient.request"
TAG = Tag(
    name="my_tag",
    allowed_values=["value1", "value2", "value3"],
    comment="Test tag",
    propagate="ON_DEPENDENCY",
    on_conflict="ALLOWED_VALUES_SEQUENCE",
)


def test_create_tag(fake_root, tags):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/schemas/my_schema/tags")
    kwargs = extra_params(
        query_params=[],
        body={
            "allowed_values": ["value1", "value2", "value3"],
            "comment": "Test tag",
            "name": "my_tag",
            "propagate": "ON_DEPENDENCY",
            "on_conflict": "ALLOWED_VALUES_SEQUENCE",
        },
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        tag_res = tags.create(TAG)
        assert isinstance(tag_res, TagResource)
        assert tag_res.name == "my_tag"
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = tags.create_async(TAG)
        assert isinstance(op, PollingOperation)
        tag_res = op.result()
        assert tag_res.name == "my_tag"
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_iter_tag(fake_root, tags):
    args = (fake_root, "GET", BASE_URL + "/databases/my_db/schemas/my_schema/tags")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        tags.iter()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        op = tags.iter_async()
        assert isinstance(op, PollingOperation)
        it = op.result()
        assert list(it) == []
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_fetch_tag(fake_root, tag):
    args = (fake_root, "GET", BASE_URL + "/databases/my_db/schemas/my_schema/tags/my_tag")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response(TAG.to_json())
        tag.fetch()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response(TAG.to_json())
        op = tag.fetch_async()
        assert isinstance(op, PollingOperation)
        fetched_tag = op.result()
        assert fetched_tag.to_dict() == TAG.to_dict()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_drop_tag(fake_root, tag):
    args = (fake_root, "DELETE", BASE_URL + "/databases/my_db/schemas/my_schema/tags/my_tag")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        tag.drop()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = tag.drop_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_undrop_tag(fake_root, tag):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/schemas/my_schema/tags/my_tag:undrop")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        tag.undrop()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = tag.undrop_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_create_or_alter_tag(fake_root, tag):
    args = (fake_root, "PUT", BASE_URL + "/databases/my_db/schemas/my_schema/tags/my_tag")
    kwargs = extra_params(
        query_params=[],
        body={
            "allowed_values": ["value1", "value2", "value3"],
            "comment": "Test tag",
            "name": "my_tag",
            "propagate": "ON_DEPENDENCY",
            "on_conflict": "ALLOWED_VALUES_SEQUENCE",
        },
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        tag.create_or_alter(TAG)
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = tag.create_or_alter_async(TAG)
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_rename_tag(fake_root, tag, tags):
    args = (
        fake_root,
        "POST",
        BASE_URL + "/databases/my_db/schemas/my_schema/tags/my_tag:rename?targetName=new_tag",
    )
    kwargs = extra_params(query_params=[("targetName", "new_tag")])

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        tag.rename("new_tag")
        assert tag.name == "new_tag"
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        tag_res = tags["another_tag"]
        op = tag_res.rename_async("new_tag")
        assert isinstance(op, PollingOperation)
        op.result()
        assert tag_res.name == "new_tag"
    args2 = (
        fake_root,
        "POST",
        BASE_URL + "/databases/my_db/schemas/my_schema/tags/another_tag:rename?targetName=new_tag",
    )
    mocked_request.assert_called_once_with(*args2, **kwargs)


def test_rename_tag_with_options(fake_root, tag):
    args = (
        fake_root,
        "POST",
        BASE_URL
        + "/databases/my_db/schemas/my_schema/tags/my_tag:rename?ifExists=True&targetDatabase=other_db&targetSchema=other_schema&targetName=new_tag",
    )
    kwargs = extra_params(
        query_params=[
            ("ifExists", True),
            ("targetDatabase", "other_db"),
            ("targetSchema", "other_schema"),
            ("targetName", "new_tag"),
        ]
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        tag.rename("new_tag", target_database="other_db", target_schema="other_schema", if_exists=True)
        assert tag.name == "new_tag"
    mocked_request.assert_called_once_with(*args, **kwargs)



"""Manages Snowflake Views."""

from ..view._generated.models import View, ViewColumn
from ._view import ViewCollection, ViewResource


__all__ = ["View", "ViewResource", "ViewCollection", "ViewColumn"]



from collections.abc import Iterator
from typing import TYPE_CHECKING, Optional

from snowflake.core._internal.telemetry import api_telemetry
from snowflake.core._operation import PollingOperation

from . import CatalogIntegration
from ._generated.api.catalog_integration_api_base import (
    CatalogIntegrationCollectionBase,
    CatalogIntegrationResourceBase,
)


if TYPE_CHECKING:
    from snowflake.core._root import Root


class CatalogIntegrationCollection(CatalogIntegrationCollectionBase):
    """Represents the collection operations on the Snowflake Catalog Integration resource.

    With this collection, you can create, iterate through, and fetch catalog integrations
    that you have access to in the current context.

    Examples
    ________
    Creating a catalog integration instance by glue:

    >>> root.catalog_integrations.create(
    ...     CatalogIntegration(
    ...         name="my_catalog_integration",
    ...         catalog=Glue(
    ...             catalog_namespace="abcd-ns",
    ...             glue_aws_role_arn="arn:aws:iam::123456789012:role/sqsAccess",
    ...             glue_catalog_id="1234567",
    ...         ),
    ...         table_format="ICEBERG",
    ...         enabled=True,
    ...     )
    ... )

    Creating a catalog integration instance by object store:

    >>> root.catalog_integrations.create(
    ...     CatalogIntegration(
    ...         name="my_catalog_integration",
    ...         catalog=ObjectStore(),
    ...         table_format="ICEBERG",
    ...         enabled=True,
    ...     )
    ... )

    Creating a catalog integration instance by polaris:

    >>> root.catalog_integrations.create(
    ...     CatalogIntegration(
    ...         name="my_catalog_integration",
    ...         catalog=Polaris(
    ...             catalog_namespace="abcd-ns",
    ...             rest_config=RestConfig(
    ...                 catalog_uri="https://my_account.snowflakecomputing.com/polaris/api/catalog",
    ...                 warehouse_name="my_warehouse",
    ...             ),
    ...             rest_authenticator=OAuth(
    ...                 type="OAUTH",
    ...                 oauth_client_id="my_oauth_client_id",
    ...                 oauth_client_secret="my_oauth_client_secret",
    ...                 oauth_allowed_scopes=["PRINCIPAL_ROLE:ALL"],
    ...             ),
    ...         ),
    ...         table_format="ICEBERG",
    ...         enabled=True,
    ...     )
    ... )
    """

    def __init__(self, root: "Root"):
        super().__init__(root, CatalogIntegrationResource)

    @api_telemetry
    def iter(self, like: Optional[str] = None) -> Iterator[CatalogIntegration]:
        """Iterate through ``CatalogIntegration`` objects from Snowflake.

        Parameters
        __________
        like: str, optional
            A case-insensitive string functioning as a filter, with support for SQL
            wildcard characters (% and _).
        """
        return super().iter(like=like)

    @api_telemetry
    def iter_async(self, like: Optional[str] = None) -> PollingOperation[Iterator[CatalogIntegration]]:
        """An asynchronous version of :func:`iter`.

        Refer to :class:`~snowflake.core.PollingOperation` for more information on asynchronous execution and
        the return type.
        """  # noqa: D401
        return super().iter_async(like=like)


class CatalogIntegrationResource(CatalogIntegrationResourceBase):
    """Represents a reference to a Snowflake Catalog Integration resource.

    With this catalog integration reference, you can fetch information about a catalog integration,
    as well as perform certain actions on it.
    """

    _plural_name = "catalog_integrations"



"""
Cortex Agent API.

OpenAPI 3.0 specification for the Cortex Agent API  # noqa: E501

The version of the OpenAPI document: 0.1.0
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, List, Optional

from pydantic import BaseModel, ConfigDict, StrictStr

from snowflake.core.cortex.lite_agent_service._generated.models.sample_question import (
    SampleQuestion,
    SampleQuestionModel,
)


class AgentInstructions(BaseModel):
    """A model object representing the AgentInstructions resource.

    Constructs an object of type AgentInstructions with the provided properties.

    Parameters
    __________
    response : str, optional
        Instructions for response generation.
    orchestration : str, optional
        Instructions for orchestrating different tools.
    system : str, optional
        System instructions for the agent.
    sample_questions : list[SampleQuestion], optional
        Sample questions and optional answers for the agent.
    """

    response: Optional[StrictStr] = None

    orchestration: Optional[StrictStr] = None

    system: Optional[StrictStr] = None

    sample_questions: Optional[List[SampleQuestion]] = None

    __properties = ["response", "orchestration", "system", "sample_questions"]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> AgentInstructions:
        """Create an instance of AgentInstructions from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        # override the default output from pydantic by calling `to_dict()` of each item in sample_questions (list)
        _items = []
        if self.sample_questions:
            for _item in self.sample_questions:
                if _item:
                    _items.append(_item.to_dict())
            _dict["sample_questions"] = _items

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> AgentInstructions:
        """Create an instance of AgentInstructions from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return AgentInstructions.model_validate(obj)

        _obj = AgentInstructions.model_validate(
            {
                "response": obj.get("response"),
                "orchestration": obj.get("orchestration"),
                "system": obj.get("system"),
                "sample_questions": [SampleQuestion.from_dict(_item) for _item in obj.get("sample_questions")]
                if obj.get("sample_questions") is not None
                else None,
            }
        )

        return _obj


class AgentInstructionsModel:
    def __init__(
        self,  # optional properties
        response: Optional[str] = None,
        orchestration: Optional[str] = None,
        system: Optional[str] = None,
        sample_questions: Optional[list[SampleQuestion]] = None,
    ):
        """A model object representing the AgentInstructions resource.

        Constructs an object of type AgentInstructions with the provided properties.

        Parameters
        __________
        response : str, optional
            Instructions for response generation.

        orchestration : str, optional
            Instructions for orchestrating different tools.

        system : str, optional
            System instructions for the agent.

        sample_questions : list[SampleQuestion], optional
            Sample questions and optional answers for the agent.
        """
        self.response = response
        self.orchestration = orchestration
        self.system = system
        self.sample_questions = sample_questions

    __properties = ["response", "orchestration", "system", "sample_questions"]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return AgentInstructions(
            response=self.response,
            orchestration=self.orchestration,
            system=self.system,
            sample_questions=[x._to_model() for x in self.sample_questions]
            if self.sample_questions is not None
            else None,
        )

    @classmethod
    def _from_model(cls, model) -> AgentInstructionsModel:
        return AgentInstructionsModel(
            response=model.response,
            orchestration=model.orchestration,
            system=model.system,
            sample_questions=[SampleQuestionModel._from_model(x) for x in model.sample_questions]
            if model.sample_questions
            else None,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a AgentInstructions.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the AgentInstructions object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> AgentInstructionsModel:
        """Create an instance of AgentInstructions from a dict.

        This method constructs a AgentInstructions object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        AgentInstructions
            A AgentInstructions object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(AgentInstructions.from_dict(obj))


AgentInstructions._model_class = AgentInstructionsModel



from contextlib import suppress

import pytest as pytest

from snowflake.core import CreateMode
from snowflake.core.exceptions import NotFoundError
from snowflake.core.notebook import Notebook
from tests.integ.notebook.conftest import notebook_file
from tests.integ.utils import random_string


@pytest.mark.min_sf_ver("8.37.0")
def test_create_and_fetch(notebooks, notebook_stage_with_file, warehouse):
    notebook_name = random_string(10, "test_notebook_")
    notebook_handle = notebooks[notebook_name]

    nb = Notebook(
        name=notebook_name,
        query_warehouse=warehouse.name,
        from_location=f"@{notebook_stage_with_file.name}",
        main_file=notebook_file,
    )

    notebook = notebooks.create(nb)
    notebooks.create(nb, mode=CreateMode.or_replace)

    try:
        fetch_handle = notebook_handle.fetch()
        assert fetch_handle.name.upper() == notebook_name.upper()

        # This returns None, so it's commented out now; have to confirm if this is an actual bug.
        # assert fetch_handle.query_warehouse.name.upper() == warehouse.name.upper()

        assert fetch_handle.from_location.upper() == f"@{notebook_stage_with_file.name}".upper()
        assert fetch_handle.main_file.upper() == notebook_file.upper()

    finally:
        with suppress(NotFoundError):
            notebook.drop()



import os
import types

from decimal import Decimal
from ssl import CERT_REQUIRED
from unittest import mock

import pytest

import snowflake.core._http_requests as http


@pytest.mark.parametrize(
    ("inputs", "expected_output"),
    (
        # Simplest case
        (("simple_url", {}, {}, ""), "simple_url"),
        # Embedding case
        (("databases/{database}", {"database": "asd"}, {}, ""), "databases/asd"),
        # Collections formats
        (
            ("items/{items}", {"items": ["bread", "butter", "cheese", "cold_cuts"]}, {"items": "csv"}, ""),
            "items/bread%2Cbutter%2Ccheese%2Ccold_cuts",
        ),
        # Safe quoting (same as last one, but don't change ',' into '%2C')
        (
            ("items/{items}", {"items": ["bread", "butter", "cheese", "cold_cuts"]}, {"items": "csv"}, ",/"),
            "items/bread,butter,cheese,cold_cuts",
        ),
    ),
)
def test_resolve_url(inputs, expected_output):
    assert http.resolve_url(*inputs) == expected_output


@pytest.fixture(autouse=True)
def _reset_connection_pool():
    # Ensure singleton pool does not leak across tests
    original = http.CONNECTION_POOL
    http.CONNECTION_POOL = None
    try:
        yield
    finally:
        http.CONNECTION_POOL = original


@mock.patch.dict(os.environ, {"HTTPS_PROXY": ""}, clear=False)
@mock.patch.object(http.urllib3, "ProxyManager")
@mock.patch.object(http.urllib3, "PoolManager")
def test_create_connection_pool_uses_proxy_from_configuration(pool_mock, proxy_mock):
    configuration = types.SimpleNamespace(
        verify_ssl=True,
        ssl_ca_cert=None,
        cert_file=None,
        key_file=None,
        assert_hostname=None,
        retries=None,
        socket_options=None,
        connection_pool_maxsize=None,
        proxy="https://proxy.local:8443",
        proxy_headers={"X-Proxy": "1"},
    )

    pool = http.create_connection_pool(configuration)
    assert isinstance(pool, http.SFPoolManager)
    proxy_mock.assert_called_once_with(
        num_pools=4,
        maxsize=4,
        cert_reqs=CERT_REQUIRED,
        ca_certs=None,
        cert_file=None,
        key_file=None,
        proxy_url="https://proxy.local:8443",
        proxy_headers={"X-Proxy": "1"},
    )
    pool_mock.assert_not_called()


@mock.patch.dict(os.environ, {"HTTPS_PROXY": "https://env-proxy:3128"}, clear=False)
@mock.patch.object(http.urllib3, "ProxyManager")
@mock.patch.object(http.urllib3, "PoolManager")
def test_create_connection_pool_uses_proxy_from_env(pool_mock, proxy_mock):
    configuration = types.SimpleNamespace(
        verify_ssl=True,
        ssl_ca_cert=None,
        cert_file=None,
        key_file=None,
        assert_hostname=None,
        retries=None,
        socket_options=None,
        connection_pool_maxsize=None,
        proxy=None,
        proxy_headers=None,
    )

    pool = http.create_connection_pool(configuration)
    assert isinstance(pool, http.SFPoolManager)
    proxy_mock.assert_called_once_with(
        num_pools=4,
        maxsize=4,
        cert_reqs=CERT_REQUIRED,
        ca_certs=None,
        cert_file=None,
        key_file=None,
        proxy_url="https://env-proxy:3128",
        proxy_headers={},
    )
    pool_mock.assert_not_called()


@mock.patch.dict(os.environ, {"HTTPS_PROXY": ""}, clear=False)
@mock.patch.object(http.urllib3, "ProxyManager")
@mock.patch.object(http.urllib3, "PoolManager")
def test_create_connection_pool_without_proxy(pool_mock, proxy_mock):
    configuration = types.SimpleNamespace(
        verify_ssl=True,
        ssl_ca_cert=None,
        cert_file=None,
        key_file=None,
        assert_hostname=None,
        retries=None,
        socket_options=None,
        connection_pool_maxsize=None,
        proxy=None,
        proxy_headers=None,
    )

    pool = http.create_connection_pool(configuration)
    assert isinstance(pool, http.SFPoolManager)
    proxy_mock.assert_not_called()
    pool_mock.assert_called_once_with(
        num_pools=4, maxsize=4, cert_reqs=CERT_REQUIRED, ca_certs=None, cert_file=None, key_file=None
    )


def test_sanitize_for_serialization():
    assert http.sanitize_for_serialization(Decimal("1.23")) == "1.23"



import copy
import logging
import pathlib
import tempfile

from unittest import mock

import pytest

from snowflake.core import simple_file_logging
from snowflake.core.stage import AwsCredentials, Stage


@pytest.fixture(scope="function", autouse=True)
def backup_reset_logging():
    logger = logging.getLogger("snowflake.core")
    original_level = logger.level
    original_handlers = copy.deepcopy(logger.handlers)
    try:
        yield
    finally:
        logger.level = original_level
        logger.handlers = original_handlers


def test_simple_file_logging_custom_path(tmp_path):
    temp_log_file = tmp_path / "asd.log"
    simple_file_logging(temp_log_file, level=logging.INFO)
    logger = logging.getLogger("snowflake.core")
    logger.info("simulated log message")
    logger.debug("shouldn't be in the file")
    logged_file = temp_log_file.read_text()
    assert "simulated log message" in logged_file
    assert "shouldn't be in the file" not in logged_file


def test_simple_file_logging():
    temp_log_file = pathlib.Path(tempfile.gettempdir()) / "snowflake_core.log"
    simple_file_logging()
    logger = logging.getLogger("snowflake.core")
    logger.info("simulated log message")
    logger.debug("shouldn't be in the file")
    logged_file = temp_log_file.read_text()
    assert "simulated log message" in logged_file
    assert "shouldn't be in the file" in logged_file


def test_secret_logging(caplog, stages):
    """This test makes sure that secret tokens are not logged."""
    with caplog.at_level(logging.DEBUG, "snowflake.core"):
        with mock.patch("snowflake.core._generated.api_client.ApiClient.request"):
            stages.create(
                Stage(
                    name="test_stage",
                    credentials=AwsCredentials(
                        aws_key_id="key_id_pwnd", aws_token="token_pwnd", aws_secret_key="key_pwnd"
                    ),
                )
            )
    assert "pwnd" not in caplog.text



# coding: utf-8

# flake8: noqa
"""
Snowflake Dynamic Table API.

The Snowflake Dynamic Table API is a REST API that you can use to access, update, and perform certain actions on Dynamic Tables resource in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import absolute_import

# import models into model package
from snowflake.core.dynamic_table._generated.models.downstream_lag import DownstreamLag
from snowflake.core.dynamic_table._generated.models.dynamic_table import DynamicTable
from snowflake.core.dynamic_table._generated.models.dynamic_table_clone import DynamicTableClone
from snowflake.core.dynamic_table._generated.models.dynamic_table_column import DynamicTableColumn
from snowflake.core.dynamic_table._generated.models.error_response import ErrorResponse
from snowflake.core.dynamic_table._generated.models.point_of_time import PointOfTime
from snowflake.core.dynamic_table._generated.models.point_of_time_offset import PointOfTimeOffset
from snowflake.core.dynamic_table._generated.models.point_of_time_statement import PointOfTimeStatement
from snowflake.core.dynamic_table._generated.models.point_of_time_timestamp import PointOfTimeTimestamp
from snowflake.core.dynamic_table._generated.models.success_accepted_response import SuccessAcceptedResponse
from snowflake.core.dynamic_table._generated.models.success_response import SuccessResponse
from snowflake.core.dynamic_table._generated.models.tag_assignment import TagAssignment
from snowflake.core.dynamic_table._generated.models.tag_reference import TagReference
from snowflake.core.dynamic_table._generated.models.target_lag import TargetLag
from snowflake.core.dynamic_table._generated.models.user_defined_lag import UserDefinedLag

__all__ = [
    "DownstreamLag",
    "DynamicTable",
    "DynamicTableClone",
    "DynamicTableColumn",
    "ErrorResponse",
    "PointOfTime",
    "PointOfTimeOffset",
    "PointOfTimeStatement",
    "PointOfTimeTimestamp",
    "SuccessAcceptedResponse",
    "SuccessResponse",
    "TagAssignment",
    "TagReference",
    "TargetLag",
    "UserDefinedLag",
]



"""Manages Snowflake Streams."""

from ..stream._generated.models import (
    PointOfTimeOffset,
    PointOfTimeStatement,
    PointOfTimeStream,
    PointOfTimeTimestamp,
    StreamSource,
    StreamSourceStage,
    StreamSourceTable,
    StreamSourceView,
)
from ._stream import Stream, StreamClone, StreamCollection, StreamResource


__all__ = [
    "PointOfTimeOffset",
    "PointOfTimeStatement",
    "PointOfTimeStream",
    "PointOfTimeTimestamp",
    "Stream",
    "StreamClone",
    "StreamSource",
    "StreamSourceStage",
    "StreamSourceTable",
    "StreamSourceView",
    "StreamResource",
    "StreamCollection",
]



from typing import Optional

import pytest

from snowflake.core.account import Account, AccountCollection
from tests.integ.utils import backup_role
from tests.utils import random_string


def assert_account_existence(temp_customer_account_accounts, account_name, should_exist=True):
    account_names = [account.name for account in temp_customer_account_accounts.iter()]
    if should_exist:
        assert account_name.upper() in account_names
    else:
        assert account_name.upper() not in account_names


def get_account_template():
    return Account(
        name=random_string(5, "test_create_account_"),
        comment="SNOWAPI_TEST_ACCOUNT",
        admin_name="SNOWAPI_TEST_ACCOUNT_ADMIN",
        admin_password="Password12345678",
        first_name="test_first_name",
        last_name="test_last_name",
        email="test_email",
        edition="ENTERPRISE",
        must_change_password=True,
    )


def create_account(
    temp_customer_account_accounts: AccountCollection,
    sf_cursor,
    temp_customer_account_cursor,
    account: Optional[Account] = None,
    exception: Optional[Exception] = None,
    exec_info_match: Optional[str] = None,
) -> Account:
    if account is None:
        account = get_account_template()
    blank_account_name = random_string(5, "test_create_blank_account_")
    with backup_role(sf_cursor):
        sf_cursor.execute("use role accountadmin;")
        sf_cursor.execute("alter session set QA_MODE=true;")
        sf_cursor.execute(f"CREATE ACCOUNT {blank_account_name} server_type=standard type=blank;")

        with backup_role(temp_customer_account_cursor):
            temp_customer_account_cursor.execute("use role orgadmin;")
            if exception is not None:
                with pytest.raises(exception) as exec_info:
                    temp_customer_account_accounts.create(account)
                if exec_info_match is not None:
                    assert exec_info.match(exec_info_match)
            else:
                temp_customer_account_accounts.create(account)
        sf_cursor.execute("alter session unset QA_MODE;")

    if exception is None:
        assert_account_existence(temp_customer_account_accounts, account.name.upper())
    return account


def cleanup_accounts_completely(sf_cursor):
    with backup_role(sf_cursor):
        sf_cursor.execute("use role accountadmin;")
        accounts_to_drop = sf_cursor.execute("SHOW ACCOUNTS like 'test_create_account_%';").fetchall()
        for account_to_drop in accounts_to_drop:
            sf_cursor.execute(f"DROP ACCOUNT IF EXISTS {account_to_drop[0]};")



#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#


from snowflake.core.database import DatabaseCollection


def test_collection_and_references(fake_root):
    db_collection = DatabaseCollection(fake_root)
    my_db_ref = db_collection["my_db"]
    my_schema_ref = my_db_ref.schemas["my_schema"]
    my_task_ref = my_schema_ref.tasks["my_task"]

    assert my_task_ref.name == "my_task"
    assert my_task_ref.schema.name == "my_schema"
    assert my_task_ref.database.name == "my_db"
    assert my_task_ref.schema is my_schema_ref
    assert my_task_ref.database is my_db_ref
    assert my_task_ref.collection is my_schema_ref.tasks



# coding: utf-8

# flake8: noqa

"""
Cortex Agent API.

OpenAPI 3.0 specification for the Cortex Agent API  # noqa: E501

The version of the OpenAPI document: 0.1.0
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import absolute_import

__version__ = "1.0.0"

# import apis into sdk package
from snowflake.core.cortex.lite_agent_service._generated.api.cortex_lite_agent_api import CortexLiteAgentApi

# import models into sdk package
from snowflake.core.cortex.lite_agent_service._generated.models.agent_instructions import AgentInstructions
from snowflake.core.cortex.lite_agent_service._generated.models.agent_models import AgentModels
from snowflake.core.cortex.lite_agent_service._generated.models.agent_orchestration import AgentOrchestration
from snowflake.core.cortex.lite_agent_service._generated.models.agent_orchestration_budget import (
    AgentOrchestrationBudget,
)
from snowflake.core.cortex.lite_agent_service._generated.models.agent_run_request import AgentRunRequest
from snowflake.core.cortex.lite_agent_service._generated.models.error_response import ErrorResponse
from snowflake.core.cortex.lite_agent_service._generated.models.mcp_server_spec import McpServerSpec
from snowflake.core.cortex.lite_agent_service._generated.models.mcp_server_wrapper import McpServerWrapper
from snowflake.core.cortex.lite_agent_service._generated.models.message import Message
from snowflake.core.cortex.lite_agent_service._generated.models.sample_question import SampleQuestion
from snowflake.core.cortex.lite_agent_service._generated.models.tool import Tool
from snowflake.core.cortex.lite_agent_service._generated.models.tool_choice import ToolChoice
from snowflake.core.cortex.lite_agent_service._generated.models.tool_tool_spec import ToolToolSpec
from snowflake.core.cortex.lite_agent_service._generated.models.tool_tool_spec_input_schema import (
    ToolToolSpecInputSchema,
)



from __future__ import annotations

import os

from collections.abc import Generator
from dataclasses import dataclass, field
from pathlib import Path
from subprocess import PIPE, run
from typing import cast

import pluggy
import pytest
import requests

from _pytest import runner
from _pytest.config import Config, Parser
from _pytest.main import Session
from _pytest.nodes import Item
from _pytest.reports import TestReport
from _pytest.runner import CallInfo
from _pytest.stash import StashKey
from _pytest.terminal import TerminalReporter


TEST_TYPE_OPTION = "--deflake-test-type"
PREVIOUS_OUTCOME_KEY = StashKey[dict[str, str]]()

FAILED = "failed"
FLAKY = "flaky"

APP_REPO = "snowflakedb/snowpy"
ISSUE_REPO = APP_REPO
FLAKY_LABEL = "flaky-test"
PHASES = ["setup", "call", "teardown"]


class DeflakePlugin:
    """Pytest plugin to retry tests and mark them as flaky.

    This plugin is used to determine which tests are flaky by automatically
    retrying them and marking them as "flaky" instead of "passing" if they
    pass on the retry. After the test run, a GitHub issue is opened or updated
    for each flaky test encountered.
    """

    name = "deflake"

    def __init__(self: DeflakePlugin, config: Config) -> None:
        super().__init__()
        self.exceptions: dict[str, Exception] = {}
        self.flaky_tests = 0
        self.test_run = TestRun()

        self.runner: type[runner] = config.pluginmanager.getplugin("runner")
        self.test_type: str = config.getoption(TEST_TYPE_OPTION)

        if token := os.getenv("GH_TOKEN"):
            # Grab the current commit so we can generate absolute URLs
            rev_parse = run(["git", "rev-parse", "HEAD"], text=True, stdout=PIPE)
            self.github: GitHub | None = GitHub(token, rev_parse.stdout.strip())
        else:
            self.github = None

    def pytest_sessionstart(self, session: Session) -> None:
        # The session is the root node in pytest's collection tree
        # Its path is the root directory that pytest looks in for tests, in
        # our case it's the root of the repo, so we can use it to make
        # relative paths that can be used to generate GitHub urls
        self.test_run.root = session.path

    def pytest_runtest_protocol(self, item: Item, nextitem: Item | None) -> bool:
        # The main protocol for running a single test
        # This function was adapted from pytest internals since we
        # need to override it completely to retry a test

        ihook = item.ihook

        # Call the logstart hook, which typically prints the name of the tests
        ihook.pytest_runtest_logstart(nodeid=item.nodeid, location=item.location)

        # Delegate back to the pytest runner to run all the phases for this test
        reports = self.runner.runtestprotocol(item, nextitem=nextitem)

        # Retry if the test reports that it should retry (report.should_retry is
        # set by us in pytest_runtest_makereport below)
        if any(getattr(report, "should_retry", False) for report in reports):
            self.runner.runtestprotocol(item, nextitem=nextitem)

        # Close the log line for this test
        ihook.pytest_runtest_logfinish(nodeid=item.nodeid, location=item.location)

        # Just return something so the next hook implementation isn't called
        return True

    def pytest_runtest_logstart(self, nodeid: str):
        # This is called once before each test regardless of the number of retries
        self.test_run.tests[nodeid] = TestResult(nodeid)

    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtest_makereport(self, item: Item, call: CallInfo) -> TestReport:
        # Wrap the creation of the test report for this phase to override the
        # status to "flaky" if it's a retry and it passed
        previous_outcomes = item.stash.setdefault(PREVIOUS_OUTCOME_KEY, {})

        # yield nothing, receive result when generator resumes
        result_wrapper: pluggy.Result = yield
        report = result_wrapper.get_result()

        report.should_retry = False
        if call.when in previous_outcomes:
            # This is a retry
            if previous_outcomes[call.when] == FAILED and report.outcome == "passed":
                # This test initially failed then passed on a retry, it's flaky
                report.outcome = FLAKY
                self.flaky_tests += 1
        elif report.outcome == FAILED:
            # This test should be retried
            report.should_retry = True
            # Don't count this as an error
            report.wasxfail = "Failure will be retried"

        previous_outcomes[call.when] = report.outcome
        return report

    def pytest_runtest_logreport(self, report: TestReport) -> None:
        # Record the crash data for the GitHub issue
        # and set the status of the overall test result
        test = self.test_run.tests[report.nodeid]
        phase = getattr(test, report.when)
        phase.outcome = report.outcome

        if report.outcome == FAILED:
            test.outcome = FAILED
            phase.longrepr = report.longreprtext
            if crash := getattr(report.longrepr, "reprcrash", None):
                phase.crash = Crash(path=crash.path, lineno=crash.lineno, message=crash.message)
        elif report.outcome == FLAKY:
            test.outcome = FLAKY

    @pytest.hookimpl(tryfirst=True)
    def pytest_report_teststatus(self, report: TestReport) -> tuple[str, str, str] | None:
        # Hook into the terminal reporting of the test status
        if getattr(report, "should_retry", False):
            # Don't log a status for this yet since it's not final
            return "", "", ""
        if report.outcome == FLAKY:
            # outcome category, letter used for regular output, and status used for full output, respectively
            return "flaky", "K", "FLAKY"
        # Otherwise let the default hook implementation decide the status strings
        return None

    def pytest_sessionfinish(self) -> None:
        # Called at the end of the pytest run to log flaky tests to GitHub
        for nodeid, test in self.test_run.tests.items():
            if test.outcome != FLAKY:
                continue

            if self.github and self.test_type:
                try:
                    self.github.create_or_update_flaky_test_issue(self.test_run.root, self.test_type, test)
                except Exception as e:
                    # Catch all exceptions to be logged later, don't let this fail the test run
                    self.exceptions[nodeid] = e

    def pytest_terminal_summary(self, terminalreporter: TerminalReporter) -> None:
        # Called at the end of the pytest run to print custom messages to the terminal
        lines = []
        if self.flaky_tests > 0:
            if not self.github:
                lines.append("Could not report flaky tests because GH_TOKEN was missing")
            if not self.test_type:
                lines.append(f"Could not report flaky tests because {TEST_TYPE_OPTION} was missing")
        for nodeid, e in self.exceptions.items():
            lines.append(f"Failed to create or update flaky test issue for {nodeid}: {e}")
        if lines:
            terminalreporter.write_sep("=", f"{self.name} plugin warnings", yellow=True)
            for line in lines:
                terminalreporter.write_line(line)


def pytest_configure(config: Config) -> None:
    # Register our plugin with pytest
    plugin = DeflakePlugin(config)
    config.pluginmanager.register(plugin, name=plugin.name)


def pytest_addoption(parser: Parser):
    # Add a flag so the user can tell us what kind of tests we're running
    # (used by the GitHub class to create issue titles)
    group = parser.getgroup("deflake", "deflake tests by recording flakes as GitHub issues")
    group.addoption(TEST_TYPE_OPTION, type=str, help="the type of tests being run", default="")


@dataclass(frozen=True)
class Crash:
    path: str
    lineno: int
    message: str


@dataclass
class TestPhase:
    outcome: str = "passed"
    longrepr: str = ""
    crash: Crash | None = field(default=None)


@dataclass
class TestResult:
    nodeid: str
    outcome: str = "passed"
    setup: TestPhase = field(default_factory=TestPhase)
    call: TestPhase = field(default_factory=TestPhase)
    teardown: TestPhase = field(default_factory=TestPhase)


@dataclass
class TestRun:
    root: Path = field(init=False)
    tests: dict[str, TestResult] = field(default_factory=dict)


class GitHub:
    def __init__(self, token: str, sha: str):
        self.session = requests.Session()
        self.session.headers.update({"Authorization": f"Bearer {token}"})
        self.sha = sha

    def create_or_update_flaky_test_issue(self, root: Path, test_type: str, test: TestResult) -> None:
        if existing_issue := self.get_issue(test_type, test):
            self.ensure_issue_open(existing_issue)
            self.update_issue_body(existing_issue, root, test)
            self.comment_on_issue(existing_issue, test)
        else:
            self.create_issue(root, test_type, test)

    def get_issue(self, test_type: str, test: TestResult) -> dict[str, dict] | None:
        issues = self.get(f"repos/{ISSUE_REPO}/issues", labels=FLAKY_LABEL, state="all")
        for issue in issues:
            # We prepend a JIRA ticket ID to new issues, so check for then of the title only
            if issue["title"].endswith(self.issue_title(test_type, test)):
                return issue
        return None

    def ensure_issue_open(self, issue: dict) -> None:
        if issue["state"] != "open":
            number = issue["number"]
            self.patch(f"repos/{ISSUE_REPO}/issues/{number}", state="open")

    def update_issue_body(self, issue: dict, root: Path, test: TestResult) -> None:
        number = issue["number"]
        body = self.issue_body(root, test)
        self.patch(f"repos/{ISSUE_REPO}/issues/{number}", body=body)

    def comment_on_issue(self, issue: dict, test: TestResult) -> dict:
        body = "+1"
        for phase in PHASES:
            if getattr(test, phase).outcome in (FAILED, FLAKY):
                body += f"\n```python\n{getattr(test, phase).crash.message}\n```\n"
                break

        number = issue["number"]
        return cast(dict, self.post(f"repos/{ISSUE_REPO}/issues/{number}/comments", body=body))

    def create_issue(self, root: Path, test_type: str, test: TestResult) -> dict:
        title = self.issue_title(test_type, test)
        body = self.issue_body(root, test)
        return cast(dict, self.post(f"repos/{ISSUE_REPO}/issues", title=title, body=body, labels=[FLAKY_LABEL]))

    def issue_title(self, test_type: str, test: TestResult) -> str:
        test_name = self.nodeid_function_name(test.nodeid)
        return f"Flaky {test_type.lower()} test: `{test_name}`"

    def issue_body(self, root: Path, test: TestResult) -> str:
        return "\n".join(
            [f"Nodeid: `{test.nodeid}`"]
            + [line for phase in PHASES for line in self.phase_lines(root, phase, getattr(test, phase))]
        )

    def phase_lines(self, root: Path, phase: str, phase_info: TestPhase | None) -> Generator[str, None, None]:
        if phase_info is None:
            return

        if phase_info.outcome == "passed":
            yield f"# 🟢 {phase.title()} passed"
            return

        yield f"# 🔴 {phase.title()} failed"
        yield f"```python\n{phase_info.longrepr}\n```"

        if phase_info.crash:
            crash_info = phase_info.crash
            path = Path(crash_info.path).relative_to(root).as_posix()
            lineno = crash_info.lineno
            url = f"https://github.com/{APP_REPO}/blob/{self.sha}/{path}#L{lineno}"
            yield url

    @staticmethod
    def nodeid_function_name(nodeid: str) -> str:
        return nodeid.split("::", 1)[-1]

    def get(self, path: str, **params) -> dict | list:
        return self.api("get", path, params=params)

    def post(self, path: str, **data) -> dict | list:
        return self.api("post", path, json=data)

    def patch(self, path: str, **data) -> dict | list:
        return self.api("patch", path, json=data)

    def api(self, method, path, *args, **kwargs) -> dict | list:
        assert self.session, "Github API must be configured"
        url = f"https://api.github.com/{path}"
        resp = self.session.request(method, url, *args, **kwargs)
        resp.raise_for_status()
        return resp.json()



import pytest

from snowflake.core.exceptions import NotFoundError
from snowflake.core.user_defined_function import PythonFunction, ReturnDataType, UserDefinedFunction
from tests.utils import random_string


pytestmark = [pytest.mark.min_sf_ver("8.38.0"), pytest.mark.usefixtures("anaconda_package_available")]


def test_drop(user_defined_functions):
    user_defined_function_name = random_string(10, "test_drop_user_defined_function_")

    user_defined_function_handle = user_defined_functions.create(
        UserDefinedFunction(
            name=user_defined_function_name,
            arguments=[],
            return_type=ReturnDataType(datatype="VARIANT"),
            language_config=PythonFunction(runtime_version="3.13", packages=[], handler="udf"),
            body="""
def udf():
    return {"key": "value"}
            """,
        )
    )

    user_defined_function_handle.fetch()

    user_defined_function_handle.drop()

    with pytest.raises(NotFoundError):
        user_defined_function_handle.fetch()

    with pytest.raises(NotFoundError):
        user_defined_function_handle.drop()

    user_defined_function_handle.drop(if_exists=True)



import copy

import pytest

from snowflake.core._internal.utils import normalize_and_unquote_name
from snowflake.core.exceptions import ConflictError
from snowflake.core.stage import Stage, StageCollection
from tests.utils import random_string


def test_create_stage(stages: StageCollection):
    new_stage_def = Stage(name=random_string(10, "test_stage_int_test_"))
    new_stage_def.comment = "stage first"
    try:
        stage = stages.create(new_stage_def)
        created_stage = stage.fetch()
        assert created_stage.name == new_stage_def.name.upper()
        assert created_stage.comment == new_stage_def.comment

        with pytest.raises(ConflictError):
            stages.create(new_stage_def, mode="error_if_exists")

        new_stage_def_1 = copy.deepcopy(new_stage_def)
        new_stage_def_1.comment = "stage second"
        stage = stages.create(new_stage_def_1, mode="if_not_exists")

        created_stage = stage.fetch()
        assert created_stage.name == new_stage_def.name.upper()
        assert created_stage.comment == new_stage_def.comment
    finally:
        stage.drop()

    try:
        stage = stages.create(new_stage_def_1, mode="or_replace")

        created_stage = stage.fetch()
        assert created_stage.name == new_stage_def_1.name.upper()
        assert created_stage.comment == new_stage_def_1.comment
    finally:
        stage.drop()

    try:
        stage_name = random_string(10, "test_stage_INT_test_")
        stage_name_case_sensitive = '"' + stage_name + '"'
        new_stage_def = Stage(name=stage_name_case_sensitive)
        stage = stages.create(new_stage_def)

        created_stage = stage.fetch()
        assert created_stage.name == normalize_and_unquote_name(new_stage_def.name)
    finally:
        stage.drop()



from typing import TYPE_CHECKING

from ._generated.api.streamlit_api_base import StreamlitCollectionBase, StreamlitResourceBase


if TYPE_CHECKING:
    from snowflake.core.schema import SchemaResource


class StreamlitCollection(StreamlitCollectionBase):
    """Represents the collection operations on the Snowflake Streamlit resource."""

    def __init__(self, schema: "SchemaResource"):
        super().__init__(schema, StreamlitResource)


class StreamlitResource(StreamlitResourceBase):
    """Represents a reference to a Snowflake Streamlit."""

    _plural_name = "streamlits"



"""
Snowflake Notebook API.

The Snowflake Notebook API is a REST API that you can use to access, update, and perform certain actions on Notebook resource in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re

from datetime import datetime
from typing import Any, List, Optional

from pydantic import BaseModel, ConfigDict, Field, StrictInt, StrictStr, field_validator
from typing_extensions import Annotated

from snowflake.core.notebook._generated.models.version_details import VersionDetails, VersionDetailsModel


class Notebook(BaseModel):
    """A model object representing the Notebook resource.

    Constructs an object of type Notebook with the provided properties.

    Parameters
    __________
    name : str
        Name of the Notebook
    version : str, optional
        User specified version alias
    from_location : str, optional
        Location to copy the file from. This must be a Snowflake stage location.
    main_file : str, optional
        Name + path of the file for the Notebook
    comment : str, optional
        user comment associated to an object in the dictionary
    default_version : str, optional
        The default version name of a file based entity.
    query_warehouse : str, optional
        Warehouse against which the queries issued by the Notebook app are run against
    created_on : datetime, optional
        Date and time when the Notebook was created — **Read-only:** *any user-provided value will be ignored.*
    database_name : str, optional
        Database in which the Notebook is stored — **Read-only:** *any user-provided value will be ignored.*
    schema_name : str, optional
        Schema in which the Notebook is stored — **Read-only:** *any user-provided value will be ignored.*
    owner : str, optional
        Role that owns the Notebook — **Read-only:** *any user-provided value will be ignored.*
    owner_role_type : str, optional
        The type of role that owns the Notebook — **Read-only:** *any user-provided value will be ignored.*
    url_id : str, optional
        A unique ID associated with the Notebook object's URL — **Read-only:** *any user-provided value will be ignored.*
    title : str, optional
        User facing title of the Notebook app
    default_packages : str, optional
        Default packages for Notebook apps — **Read-only:** *any user-provided value will be ignored.*
    user_packages : str, optional
        User packages of the Notebook — **Read-only:** *any user-provided value will be ignored.*
    runtime_name : str, optional
        The runtime to run the Notebook on. If this is not set, the Warehouse Runtime is assumed
    compute_pool : str, optional
        Compute pool name where the Notebook runs. Required when runtime_name is set to SYSTEM$BASIC_RUNTIME or SYSTEM$GPU_RUNTIME
    import_urls : list[str], optional
        List of files to be imported from a stage.
    external_access_integrations : list[str], optional
        List of external access integrations attached to this Notebook
    external_access_secrets : str, optional
        Secrets to be used with this function for external access — **Read-only:** *any user-provided value will be ignored.*
    idle_auto_shutdown_time_seconds : int, optional
        Sets the time in seconds for when to shutdown an idle Notebook.
    default_version_details : VersionDetails, optional

    last_version_details : VersionDetails, optional

    live_version_location_uri : str, optional
        The current version location — **Read-only:** *any user-provided value will be ignored.*
    budget : str, optional
        Name of the budget if the Notebook is monitored by a budget — **Read-only:** *any user-provided value will be ignored.*
    """

    name: Annotated[str, Field(strict=True)]

    version: Optional[StrictStr] = None

    from_location: Optional[StrictStr] = Field(default=None, alias="fromLocation")

    main_file: Optional[StrictStr] = None

    comment: Optional[StrictStr] = None

    default_version: Optional[StrictStr] = None

    query_warehouse: Optional[StrictStr] = None

    created_on: Optional[datetime] = None

    database_name: Optional[Annotated[str, Field(strict=True)]] = None

    schema_name: Optional[Annotated[str, Field(strict=True)]] = None

    owner: Optional[Annotated[str, Field(strict=True)]] = None

    owner_role_type: Optional[Annotated[str, Field(strict=True)]] = None

    url_id: Optional[StrictStr] = None

    title: Optional[StrictStr] = None

    default_packages: Optional[StrictStr] = None

    user_packages: Optional[StrictStr] = None

    runtime_name: Optional[StrictStr] = None

    compute_pool: Optional[StrictStr] = None

    import_urls: Optional[List[StrictStr]] = None

    external_access_integrations: Optional[List[StrictStr]] = None

    external_access_secrets: Optional[StrictStr] = None

    idle_auto_shutdown_time_seconds: Optional[StrictInt] = None

    default_version_details: Optional[VersionDetails] = None

    last_version_details: Optional[VersionDetails] = None

    live_version_location_uri: Optional[StrictStr] = None

    budget: Optional[Annotated[str, Field(strict=True)]] = None

    __properties = [
        "name",
        "version",
        "fromLocation",
        "main_file",
        "comment",
        "default_version",
        "query_warehouse",
        "created_on",
        "database_name",
        "schema_name",
        "owner",
        "owner_role_type",
        "url_id",
        "title",
        "default_packages",
        "user_packages",
        "runtime_name",
        "compute_pool",
        "import_urls",
        "external_access_integrations",
        "external_access_secrets",
        "idle_auto_shutdown_time_seconds",
        "default_version_details",
        "last_version_details",
        "live_version_location_uri",
        "budget",
    ]

    @field_validator("name")
    def name_validate_regular_expression(cls, v):
        if not re.match(r"""^\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*$""", v):
            raise ValueError(r"""must validate the regular expression /^"([^"]|"")+"|[a-zA-Z_][a-zA-Z0-9_$]*$/""")
        return v

    @field_validator("database_name")
    def database_name_validate_regular_expression(cls, v):
        if v is None:
            return v
        if not re.match(r"""^\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*$""", v):
            raise ValueError(r"""must validate the regular expression /^"([^"]|"")+"|[a-zA-Z_][a-zA-Z0-9_$]*$/""")
        return v

    @field_validator("schema_name")
    def schema_name_validate_regular_expression(cls, v):
        if v is None:
            return v
        if not re.match(r"""^\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*$""", v):
            raise ValueError(r"""must validate the regular expression /^"([^"]|"")+"|[a-zA-Z_][a-zA-Z0-9_$]*$/""")
        return v

    @field_validator("owner")
    def owner_validate_regular_expression(cls, v):
        if v is None:
            return v
        if not re.match(r"""^\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*$""", v):
            raise ValueError(r"""must validate the regular expression /^"([^"]|"")+"|[a-zA-Z_][a-zA-Z0-9_$]*$/""")
        return v

    @field_validator("owner_role_type")
    def owner_role_type_validate_regular_expression(cls, v):
        if v is None:
            return v
        if not re.match(r"""^\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*$""", v):
            raise ValueError(r"""must validate the regular expression /^"([^"]|"")+"|[a-zA-Z_][a-zA-Z0-9_$]*$/""")
        return v

    @field_validator("budget")
    def budget_validate_regular_expression(cls, v):
        if v is None:
            return v
        if not re.match(r"""^\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*$""", v):
            raise ValueError(r"""must validate the regular expression /^"([^"]|"")+"|[a-zA-Z_][a-zA-Z0-9_$]*$/""")
        return v

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Notebook:
        """Create an instance of Notebook from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update(
                {
                    "created_on",
                    "database_name",
                    "schema_name",
                    "owner",
                    "owner_role_type",
                    "url_id",
                    "default_packages",
                    "user_packages",
                    "external_access_secrets",
                    "live_version_location_uri",
                    "budget",
                }
            )

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        # override the default output from pydantic by calling `to_dict()` of default_version_details
        if self.default_version_details:
            _dict["default_version_details"] = self.default_version_details.to_dict()

        # override the default output from pydantic by calling `to_dict()` of last_version_details
        if self.last_version_details:
            _dict["last_version_details"] = self.last_version_details.to_dict()

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> Notebook:
        """Create an instance of Notebook from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return Notebook.model_validate(obj)

        _obj = Notebook.model_validate(
            {
                "name": obj.get("name"),
                "version": obj.get("version"),
                "from_location": obj.get("fromLocation"),
                "main_file": obj.get("main_file"),
                "comment": obj.get("comment"),
                "default_version": obj.get("default_version"),
                "query_warehouse": obj.get("query_warehouse"),
                "created_on": obj.get("created_on"),
                "database_name": obj.get("database_name"),
                "schema_name": obj.get("schema_name"),
                "owner": obj.get("owner"),
                "owner_role_type": obj.get("owner_role_type"),
                "url_id": obj.get("url_id"),
                "title": obj.get("title"),
                "default_packages": obj.get("default_packages"),
                "user_packages": obj.get("user_packages"),
                "runtime_name": obj.get("runtime_name"),
                "compute_pool": obj.get("compute_pool"),
                "import_urls": obj.get("import_urls"),
                "external_access_integrations": obj.get("external_access_integrations"),
                "external_access_secrets": obj.get("external_access_secrets"),
                "idle_auto_shutdown_time_seconds": obj.get("idle_auto_shutdown_time_seconds"),
                "default_version_details": VersionDetails.from_dict(obj.get("default_version_details"))
                if obj.get("default_version_details") is not None
                else None,
                "last_version_details": VersionDetails.from_dict(obj.get("last_version_details"))
                if obj.get("last_version_details") is not None
                else None,
                "live_version_location_uri": obj.get("live_version_location_uri"),
                "budget": obj.get("budget"),
            }
        )

        return _obj


class NotebookModel:
    def __init__(
        self,
        name: str,
        # optional properties
        version: Optional[str] = None,
        from_location: Optional[str] = None,
        main_file: Optional[str] = None,
        comment: Optional[str] = None,
        default_version: Optional[str] = None,
        query_warehouse: Optional[str] = None,
        created_on: Optional[datetime] = None,
        database_name: Optional[str] = None,
        schema_name: Optional[str] = None,
        owner: Optional[str] = None,
        owner_role_type: Optional[str] = None,
        url_id: Optional[str] = None,
        title: Optional[str] = None,
        default_packages: Optional[str] = None,
        user_packages: Optional[str] = None,
        runtime_name: Optional[str] = None,
        compute_pool: Optional[str] = None,
        import_urls: Optional[list[str]] = None,
        external_access_integrations: Optional[list[str]] = None,
        external_access_secrets: Optional[str] = None,
        idle_auto_shutdown_time_seconds: Optional[int] = None,
        default_version_details: Optional[VersionDetails] = None,
        last_version_details: Optional[VersionDetails] = None,
        live_version_location_uri: Optional[str] = None,
        budget: Optional[str] = None,
    ):
        """A model object representing the Notebook resource.

        Constructs an object of type Notebook with the provided properties.

        Parameters
        __________
        name : str
            Name of the Notebook
        version : str, optional
            User specified version alias
        from_location : str, optional
            Location to copy the file from. This must be a Snowflake stage location.
        main_file : str, optional
            Name + path of the file for the Notebook
        comment : str, optional
            user comment associated to an object in the dictionary
        default_version : str, optional
            The default version name of a file based entity.
        query_warehouse : str, optional
            Warehouse against which the queries issued by the Notebook app are run against
        created_on : datetime, optional
            Date and time when the Notebook was created.
        database_name : str, optional
            Database in which the Notebook is stored
        schema_name : str, optional
            Schema in which the Notebook is stored
        owner : str, optional
            Role that owns the Notebook
        owner_role_type : str, optional
            The type of role that owns the Notebook
        url_id : str, optional
            A unique ID associated with the Notebook object's URL.
        title : str, optional
            User facing title of the Notebook app
        default_packages : str, optional
            Default packages for Notebook apps.
        user_packages : str, optional
            User packages of the Notebook
        runtime_name : str, optional
            The runtime to run the Notebook on. If this is not set, the Warehouse Runtime is assumed
        compute_pool : str, optional
            Compute pool name where the Notebook runs. Required when runtime_name is set to SYSTEM$BASIC_RUNTIME or SYSTEM$GPU_RUNTIME
        import_urls : list[str], optional
            List of files to be imported from a stage.
        external_access_integrations : list[str], optional
            List of external access integrations attached to this Notebook
        external_access_secrets : str, optional
            Secrets to be used with this function for external access
        idle_auto_shutdown_time_seconds : int, optional
            Sets the time in seconds for when to shutdown an idle Notebook.
        default_version_details : VersionDetails, optional

        last_version_details : VersionDetails, optional

        live_version_location_uri : str, optional
            The current version location
        budget : str, optional
            Name of the budget if the Notebook is monitored by a budget
        """
        self.name = name
        self.version = version
        self.from_location = from_location
        self.main_file = main_file
        self.comment = comment
        self.default_version = default_version
        self.query_warehouse = query_warehouse
        self.created_on = created_on
        self.database_name = database_name
        self.schema_name = schema_name
        self.owner = owner
        self.owner_role_type = owner_role_type
        self.url_id = url_id
        self.title = title
        self.default_packages = default_packages
        self.user_packages = user_packages
        self.runtime_name = runtime_name
        self.compute_pool = compute_pool
        self.import_urls = import_urls
        self.external_access_integrations = external_access_integrations
        self.external_access_secrets = external_access_secrets
        self.idle_auto_shutdown_time_seconds = idle_auto_shutdown_time_seconds
        self.default_version_details = default_version_details
        self.last_version_details = last_version_details
        self.live_version_location_uri = live_version_location_uri
        self.budget = budget

    __properties = [
        "name",
        "version",
        "fromLocation",
        "main_file",
        "comment",
        "default_version",
        "query_warehouse",
        "created_on",
        "database_name",
        "schema_name",
        "owner",
        "owner_role_type",
        "url_id",
        "title",
        "default_packages",
        "user_packages",
        "runtime_name",
        "compute_pool",
        "import_urls",
        "external_access_integrations",
        "external_access_secrets",
        "idle_auto_shutdown_time_seconds",
        "default_version_details",
        "last_version_details",
        "live_version_location_uri",
        "budget",
    ]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return Notebook(
            name=self.name,
            version=self.version,
            from_location=self.from_location,
            main_file=self.main_file,
            comment=self.comment,
            default_version=self.default_version,
            query_warehouse=self.query_warehouse,
            created_on=self.created_on,
            database_name=self.database_name,
            schema_name=self.schema_name,
            owner=self.owner,
            owner_role_type=self.owner_role_type,
            url_id=self.url_id,
            title=self.title,
            default_packages=self.default_packages,
            user_packages=self.user_packages,
            runtime_name=self.runtime_name,
            compute_pool=self.compute_pool,
            import_urls=self.import_urls,
            external_access_integrations=self.external_access_integrations,
            external_access_secrets=self.external_access_secrets,
            idle_auto_shutdown_time_seconds=self.idle_auto_shutdown_time_seconds,
            default_version_details=self.default_version_details._to_model()
            if self.default_version_details is not None
            else None,
            last_version_details=self.last_version_details._to_model()
            if self.last_version_details is not None
            else None,
            live_version_location_uri=self.live_version_location_uri,
            budget=self.budget,
        )

    @classmethod
    def _from_model(cls, model) -> NotebookModel:
        return NotebookModel(
            name=model.name,
            version=model.version,
            from_location=model.from_location,
            main_file=model.main_file,
            comment=model.comment,
            default_version=model.default_version,
            query_warehouse=model.query_warehouse,
            created_on=model.created_on,
            database_name=model.database_name,
            schema_name=model.schema_name,
            owner=model.owner,
            owner_role_type=model.owner_role_type,
            url_id=model.url_id,
            title=model.title,
            default_packages=model.default_packages,
            user_packages=model.user_packages,
            runtime_name=model.runtime_name,
            compute_pool=model.compute_pool,
            import_urls=model.import_urls,
            external_access_integrations=model.external_access_integrations,
            external_access_secrets=model.external_access_secrets,
            idle_auto_shutdown_time_seconds=model.idle_auto_shutdown_time_seconds,
            default_version_details=VersionDetailsModel._from_model(model.default_version_details)
            if model.default_version_details
            else None,
            last_version_details=VersionDetailsModel._from_model(model.last_version_details)
            if model.last_version_details
            else None,
            live_version_location_uri=model.live_version_location_uri,
            budget=model.budget,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a Notebook.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the Notebook object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> NotebookModel:
        """Create an instance of Notebook from a dict.

        This method constructs a Notebook object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        Notebook
            A Notebook object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(Notebook.from_dict(obj))


Notebook._model_class = NotebookModel



from snowflake.core.cortex.chat_service._generated.models.chat_request_messages_inner import ChatRequestMessagesInner

from ._chat_service import ChatRequest, CortexChatService
from ._generated import CortexChatApi


CortexChatServiceApi = CortexChatApi

__all__ = [
    "CortexChatService",
    "ChatRequest",
    "CortexChatServiceApi",
    "ChatRequestMessagesInner",
]



from unittest import mock

import pytest

from snowflake.core import PollingOperation, Root
from snowflake.core.notebook import Notebook, NotebookResource
from snowflake.core.notebook._generated import TagAssignment, TagReference
from snowflake.core.tag import TagValue

from ...utils import BASE_URL, extra_params, mock_http_response


API_CLIENT_REQUEST = "snowflake.core._generated.api_client.ApiClient.request"
NOTEBOOK = Notebook(name="my_notebook")


@pytest.fixture
def notebooks(schema):
    return schema.notebooks


@pytest.fixture
def notebook(notebooks):
    return notebooks["my_notebook"]


def test_create_notebook(fake_root, notebooks):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/schemas/my_schema/notebooks?createMode=errorIfExists")
    kwargs = extra_params(query_params=[("createMode", "errorIfExists")], body={"name": "my_notebook"})

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        notebook_res = notebooks.create(NOTEBOOK)
        assert isinstance(notebook_res, NotebookResource)
        assert notebook_res.name == "my_notebook"
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = notebooks.create_async(NOTEBOOK)
        assert isinstance(op, PollingOperation)
        notebook_res = op.result()
        assert notebook_res.name == "my_notebook"
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_iter_notebook(fake_root, notebooks):
    args = (fake_root, "GET", BASE_URL + "/databases/my_db/schemas/my_schema/notebooks")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        notebooks.iter()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        op = notebooks.iter_async()
        assert isinstance(op, PollingOperation)
        it = op.result()
        assert list(it) == []
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_iter_notebook_with_parameters(fake_root, notebooks):
    """Test iter notebook with various filter parameters."""
    # The API uses showLimit instead of limit for the actual parameter name
    args = (
        fake_root,
        "GET",
        BASE_URL + "/databases/my_db/schemas/my_schema/notebooks?like=test"
        "%&startsWith=test&showLimit=100&fromName=start_name",
    )
    kwargs = extra_params(
        query_params=[
            ("like", "test%"),
            ("startsWith", "test"),
            ("showLimit", 100),
            ("fromName", "start_name"),
        ]
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        notebooks.iter(like="test%", starts_with="test", limit=100, from_name="start_name")
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_iter_show_limit_deprecation_warning(fake_root, notebooks):
    """Test that show_limit parameter triggers deprecation warning."""
    args = (
        fake_root,
        "GET",
        BASE_URL + "/databases/my_db/schemas/my_schema/notebooks?showLimit=10",
    )
    kwargs = extra_params(query_params=[("showLimit", 10)])

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        with pytest.warns(DeprecationWarning, match="'show_limit' is deprecated, use 'limit' instead"):
            list(notebooks.iter(show_limit=10))
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        with pytest.warns(DeprecationWarning, match="'show_limit' is deprecated, use 'limit' instead"):
            notebooks.iter_async(show_limit=10).result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_iter_limit_and_show_limit_conflict(notebooks):
    """Test that providing both limit and show_limit raises ValueError."""
    with pytest.raises(ValueError, match="Cannot specify both 'limit' and 'show_limit'"):
        list(notebooks.iter(limit=10, show_limit=5))

    with pytest.raises(ValueError, match="Cannot specify both 'limit' and 'show_limit'"):
        notebooks.iter_async(limit=10, show_limit=5).result()


def test_fetch_notebook(fake_root, notebook):
    from snowflake.core.notebook._generated.models import Notebook as NotebookModel

    model = NotebookModel(name="my_notebook")
    args = (fake_root, "GET", BASE_URL + "/databases/my_db/schemas/my_schema/notebooks/my_notebook")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response(model.to_json())
        notebook.fetch()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response(model.to_json())
        op = notebook.fetch_async()
        assert isinstance(op, PollingOperation)
        tab = op.result()
        assert tab.to_dict() == NOTEBOOK.to_dict()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_drop_notebook(fake_root, notebook):
    args = (fake_root, "DELETE", BASE_URL + "/databases/my_db/schemas/my_schema/notebooks/my_notebook")
    kwargs = extra_params(query_params=[])

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        notebook.drop()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = notebook.drop_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_drop_notebook_with_if_exists(fake_root, notebook):
    """Test drop notebook with if_exists parameter."""
    args = (
        fake_root,
        "DELETE",
        BASE_URL + "/databases/my_db/schemas/my_schema/notebooks/my_notebook?ifExists=True",
    )
    kwargs = extra_params(query_params=[("ifExists", True)])

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        notebook.drop(if_exists=True)
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = notebook.drop_async(if_exists=True)
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_rename_notebook(fake_root, notebook, notebooks):
    def format_args(notebook_name: str) -> tuple[Root, str, str]:
        return (
            fake_root,
            "POST",
            BASE_URL
            + f"/databases/my_db/schemas/my_schema/notebooks/{notebook_name}:rename?"
            + "targetName=new_notebook",
        )

    kwargs = extra_params(
        query_params=[
            ("targetName", "new_notebook"),
        ]
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        notebook.rename("new_notebook")
        assert notebook.name == "new_notebook"
    mocked_request.assert_called_once_with(*format_args("my_notebook"), **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        et_res = notebooks["another_table"]
        op = et_res.rename_async("new_notebook")
        assert isinstance(op, PollingOperation)
        op.result()
        assert et_res.name == "new_notebook"
    mocked_request.assert_called_once_with(*format_args("another_table"), **kwargs)

    assert notebooks["my_notebook"].name == "my_notebook"
    assert notebooks["new_notebook"].name == "new_notebook"


def test_rename_notebook_with_target_database_and_schema(fake_root, notebook):
    """Test rename notebook with explicit target database and schema."""
    args = (
        fake_root,
        "POST",
        BASE_URL
        + "/databases/my_db/schemas/my_schema/notebooks/my_notebook:rename?"
        + "targetDatabase=target_db&targetSchema=target_schema&targetName=new_notebook",
    )
    kwargs = extra_params(
        query_params=[
            ("targetDatabase", "target_db"),
            ("targetSchema", "target_schema"),
            ("targetName", "new_notebook"),
        ]
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        notebook.rename("new_notebook", target_database="target_db", target_schema="target_schema")
        assert notebook.name == "new_notebook"
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_rename_notebook_with_if_exists(fake_root, notebook):
    """Test rename notebook with if_exists parameter."""
    # The API puts ifExists first in the query string
    args = (
        fake_root,
        "POST",
        BASE_URL
        + "/databases/my_db/schemas/my_schema/notebooks/my_notebook:rename?"
        + "ifExists=True&targetName=new_notebook",
    )
    kwargs = extra_params(
        query_params=[
            ("ifExists", True),
            ("targetName", "new_notebook"),
        ]
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        notebook.rename("new_notebook", if_exists=True)
        assert notebook.name == "new_notebook"
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_execute_notebook(fake_root, notebook):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/schemas/my_schema/notebooks/my_notebook:execute")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        notebook.execute()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = notebook.execute_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_execute_notebook_with_async_exec(fake_root, notebook):
    """Test execute notebook with async_exec parameter."""
    args = (
        fake_root,
        "POST",
        BASE_URL + "/databases/my_db/schemas/my_schema/notebooks/my_notebook:execute?asyncExec=True",
    )
    kwargs = extra_params(query_params=[("asyncExec", True)])

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        notebook.execute(async_exec=True)
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = notebook.execute_async(async_exec=False)
        assert isinstance(op, PollingOperation)
        op.result()
    # Should have asyncExec=False in the URL
    expected_args = (
        fake_root,
        "POST",
        BASE_URL + "/databases/my_db/schemas/my_schema/notebooks/my_notebook:execute?asyncExec=False",
    )
    expected_kwargs = extra_params(query_params=[("asyncExec", False)])
    mocked_request.assert_called_once_with(*expected_args, **expected_kwargs)


def test_commit_notebook(fake_root, notebook):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/schemas/my_schema/notebooks/my_notebook:commit")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        notebook.commit()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = notebook.commit_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_add_live_version_notebook(fake_root, notebook):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/schemas/my_schema/notebooks/my_notebook:add-live-version")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        notebook.add_live_version()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = notebook.add_live_version_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_set_tags(fake_root, notebook, tag):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/schemas/my_schema/notebooks/my_notebook:set-tags")
    tags = {tag: TagValue(value="value")}
    kwargs = extra_params(
        body=[
            TagAssignment(
                tag_value=v.value, tag_name=k.name, tag_schema=k.schema.name, tag_database=k.database.name
            ).to_dict()
            for k, v in tags.items()
        ]
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        notebook.set_tags(tags)
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = notebook.set_tags_async(tags)
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_unset_tags(fake_root, notebook, tag):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/schemas/my_schema/notebooks/my_notebook:unset-tags")
    tag_resources = {tag}
    kwargs = extra_params(
        body=[
            TagReference(
                tag_name=tag_res.name, tag_schema=tag_res.schema.name, tag_database=tag_res.database.name
            ).to_dict()
            for tag_res in tag_resources
        ]
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        notebook.unset_tags(tag_resources)
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = notebook.unset_tags_async(tag_resources)
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_get_tags(fake_root, notebook):
    args = (fake_root, "GET", BASE_URL + "/databases/my_db/schemas/my_schema/notebooks/my_notebook:get-tags")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        assert notebook.get_tags() == {}
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        op = notebook.get_tags_async()
        assert isinstance(op, PollingOperation)
        assert op.result() == {}
    mocked_request.assert_called_once_with(*args, **kwargs)



from concurrent.futures import ThreadPoolExecutor

import pytest

import snowflake.core._thread_pool as thread_pool

from snowflake.core._thread_pool import get_thread_pool


@pytest.fixture(autouse=True)
def reset_pool():
    thread_pool.THREAD_POOL.reset()


def test_default_thread_pool_size():
    pool = get_thread_pool()
    assert pool._max_workers == ThreadPoolExecutor()._max_workers


def test_custom_thread_pool_size():
    pool = get_thread_pool(10)
    assert pool._max_workers == 10


def test_init_once():
    get_thread_pool(10)
    pool = get_thread_pool(20)
    assert pool._max_workers == 10



from unittest import mock
from urllib.parse import quote

import pytest

from snowflake.core import PollingOperation, Root
from snowflake.core.tag import TagValue
from snowflake.core.user_defined_function import (
    ColumnType,
    ReturnDataType,
    ReturnTable,
    SQLFunction,
    UserDefinedFunction,
    UserDefinedFunctionResource,
)
from snowflake.core.user_defined_function._generated import TagAssignment, TagReference

from ...utils import BASE_URL, extra_params, mock_http_response, random_string


API_CLIENT_REQUEST = "snowflake.core._generated.api_client.ApiClient.request"
UDF = UserDefinedFunction(
    name="my_udf", arguments=[], return_type=ReturnDataType(datatype="VARCHAR"), language_config=SQLFunction()
)


@pytest.fixture
def user_defined_functions(schema):
    return schema.user_defined_functions


@pytest.fixture
def user_defined_function(user_defined_functions):
    return user_defined_functions["my_udf()"]


def test_create_user_defined_function(fake_root, user_defined_functions):
    args = (fake_root, "POST", BASE_URL + "/databases/my_db/schemas/my_schema/user-defined-functions")
    kwargs = extra_params(
        query_params=[],
        body={
            "name": "my_udf",
            "arguments": [],
            "return_type": {"datatype": "VARCHAR", "type": "DATATYPE"},
            "language_config": {"language": "SQL"},
        },
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        udf_res = user_defined_functions.create(UDF)
        assert isinstance(udf_res, UserDefinedFunctionResource)
        assert udf_res.name_with_args == "my_udf()"
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = user_defined_functions.create_async(UDF)
        assert isinstance(op, PollingOperation)
        udf_res = op.result()
        assert udf_res.name_with_args == "my_udf()"
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_iter_user_defined_function(fake_root, user_defined_functions):
    args = (fake_root, "GET", BASE_URL + "/databases/my_db/schemas/my_schema/user-defined-functions")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        user_defined_functions.iter()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        op = user_defined_functions.iter_async()
        assert isinstance(op, PollingOperation)
        it = op.result()
        assert list(it) == []
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_fetch_user_defined_function(fake_root, user_defined_function):
    from snowflake.core.user_defined_function._generated.models import ReturnDataType as ReturnDataTypeModel
    from snowflake.core.user_defined_function._generated.models import SQLFunction as SQLFunctionModel
    from snowflake.core.user_defined_function._generated.models import UserDefinedFunction as UserDefinedFunctionModel

    model = UserDefinedFunctionModel(
        name="my_udf",
        arguments=[],
        return_type=ReturnDataTypeModel(datatype="VARCHAR"),
        language_config=SQLFunctionModel(),
    )
    args = (
        fake_root,
        "GET",
        BASE_URL + f"/databases/my_db/schemas/my_schema/user-defined-functions/{quote('my_udf()')}",
    )
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response(model.to_json())
        user_defined_function.fetch()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response(model.to_json())
        op = user_defined_function.fetch_async()
        assert isinstance(op, PollingOperation)
        tab = op.result()
        assert tab.to_dict() == UDF.to_dict()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_drop_user_defined_function(fake_root, user_defined_function):
    args = (
        fake_root,
        "DELETE",
        BASE_URL + f"/databases/my_db/schemas/my_schema/user-defined-functions/{quote('my_udf()')}",
    )
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        user_defined_function.drop()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = user_defined_function.drop_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_rename_user_defined_function(fake_root, user_defined_function, user_defined_functions):
    def format_args(udf_name: str) -> tuple[Root, str, str]:
        return (
            fake_root,
            "POST",
            BASE_URL
            + f"/databases/my_db/schemas/my_schema/user-defined-functions/{quote(udf_name)}:rename?"
            + "targetDatabase=my_db&targetSchema=my_schema&targetName=new_udf",
        )

    kwargs = extra_params(
        query_params=[("targetDatabase", "my_db"), ("targetSchema", "my_schema"), ("targetName", "new_udf")]
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        user_defined_function.rename("new_udf")
        assert user_defined_function.name_with_args == "new_udf()"
    mocked_request.assert_called_once_with(*format_args("my_udf()"), **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        udf_res2 = user_defined_functions["my_udf2()"]
        op = udf_res2.rename_async("new_udf")
        assert isinstance(op, PollingOperation)
        op.result()
        assert udf_res2.name_with_args == "new_udf()"
    mocked_request.assert_called_once_with(*format_args("my_udf2()"), **kwargs)


def test_execute_udf_table_raises_not_implemented_error(fake_root, user_defined_functions):
    udf_name = random_string(10, "test_create_user_defined_function_sql_")
    func_body = """
        SELECT 1, 2
        UNION ALL
        SELECT 3, 4
    """
    model = UserDefinedFunction(
        name=udf_name,
        arguments=[],
        return_type=ReturnTable(
            column_list=[ColumnType(name="x", datatype="INTEGER"), ColumnType(name="y", datatype="INTEGER")]
        ),
        language_config=SQLFunction(),
        body=func_body,
    )
    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        udf_handle = user_defined_functions.create(model)
        mocked_request.return_value = mock_http_response(model.to_json())
        with pytest.raises(NotImplementedError):
            udf_handle.execute([])


def test_set_tags(fake_root, user_defined_function, tag):
    args = (
        fake_root,
        "POST",
        BASE_URL + f"/databases/my_db/schemas/my_schema/user-defined-functions/{quote('my_udf()')}:set-tags",
    )
    tags = {tag: TagValue(value="value")}
    kwargs = extra_params(
        body=[
            TagAssignment(
                tag_value=v.value, tag_name=k.name, tag_schema=k.schema.name, tag_database=k.database.name
            ).to_dict()
            for k, v in tags.items()
        ]
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        user_defined_function.set_tags(tags)
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = user_defined_function.set_tags_async(tags)
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_unset_tags(fake_root, user_defined_function, tag):
    args = (
        fake_root,
        "POST",
        BASE_URL + f"/databases/my_db/schemas/my_schema/user-defined-functions/{quote('my_udf()')}:unset-tags",
    )
    tag_resources = {tag}
    kwargs = extra_params(
        body=[
            TagReference(
                tag_name=tag_res.name, tag_schema=tag_res.schema.name, tag_database=tag_res.database.name
            ).to_dict()
            for tag_res in tag_resources
        ]
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        user_defined_function.unset_tags(tag_resources)
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = user_defined_function.unset_tags_async(tag_resources)
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_get_tags(fake_root, user_defined_function):
    args = (
        fake_root,
        "GET",
        BASE_URL + f"/databases/my_db/schemas/my_schema/user-defined-functions/{quote('my_udf()')}:get-tags",
    )
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        assert user_defined_function.get_tags() == {}
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        op = user_defined_function.get_tags_async()
        assert isinstance(op, PollingOperation)
        assert op.result() == {}
    mocked_request.assert_called_once_with(*args, **kwargs)



# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
import json

from datetime import datetime
from typing import TYPE_CHECKING, Any, Optional, Union

from .._internal.utils import validate_object_name


if TYPE_CHECKING:
    from snowflake.snowpark import Session


class TaskContext:
    """Represents the context in a `Snowflake Task <https://docs.snowflake.com/en/user-guide/tasks-intro>`_.

    When a :class:`Task`'s ``definition`` is a :class:`StoredProcedureCall`, the handler of the Stored Procedure
    can use this class to set the return value of the Task so the immediate successor tasks can use it.
    """

    def __init__(self, session: "Session") -> None:
        """Initialize a :class:`TaskContext` object.

        Args:
            session: a Snowpark session.
        """
        self._session = session
        self._current_task_name: Optional[str] = None
        self._current_root_name: Optional[str] = None

    def _get_root_prefix(self) -> Optional[str]:
        root_task_name = self.get_current_root_task_name()
        root_prefix = root_task_name[root_task_name.rfind(".") + 1 :] + "$"
        current_task_name = self.get_current_task_name()
        current_task_name = current_task_name[root_task_name.rfind(".") + 1 :]
        if current_task_name.startswith(root_prefix):
            return root_prefix
        return None

    def set_return_value(self, value: Any) -> None:
        """Explicitly sets the return value for a task.

        An immediate successor task can then use :meth:`get_predecessor_return_value` to retrieve it.
        See `SYSTEM$SET_RETURN_VALUE
        <https://docs.snowflake.com/en/sql-reference/functions/system_set_return_value>`_ for details.
        This method can only be called in a Snowflake task.

        Args:
            value: The return value for a task. It will be converted to a ``str``
                when the underlying SQL system function is called.

        Example:
            >>> def task_handler(session: Session) -> None:
            >>>     from snowflake.core.task.context import TaskContext
            >>>     context = TaskContext(session)
            >>> # this return value can be retrieved by successor Tasks.
            >>>     context.set_return_value("predecessor_return_value")
        """
        self._session.call("system$set_return_value", str(value))

    def get_predecessor_return_value(self, task_name: Optional[str] = None) -> str:
        """Retrieve the return value from the predecessor task in a DAG of tasks.

        The return value is explicitly set by the predecessor task using :meth:`set_return_value`.
        This method can only be called in a Snowflake task.

        See `SYSTEM$GET_PREDECESSOR_RETURN_VALUE
        <https://docs.snowflake.com/en/sql-reference/functions/system_get_predecessor_return_value>`_ for details.

        Args:
            task_name: The task name of the predecessor task that sets the return value to be retrieved.

                - If the task has only one predecessor task that is enabled, the argument is optional.
                  If this argument is omitted, the function retrieves the return value for the only enabled predecessor
                  task.
                - If the task has multiple predecessor tasks that are enabled, this argument is required.

        Example:
            >>> def task_handler(session: Session) -> None:
            >>>     from snowflake.core.task.context import TaskContext
            >>>     context = TaskContext(session)
            >>>     pred_return_value = context.get_predecessor_return_value("pred_task_name")
        """
        if task_name:
            validate_object_name(task_name)
            root_prefix = self._get_root_prefix()
            task_name = task_name.upper()
            if root_prefix and not task_name.startswith(root_prefix):
                task_name = f"{root_prefix}{task_name}"
            return str(self._session.call("system$get_predecessor_return_value", task_name))
        return str(self._session.call("system$get_predecessor_return_value"))

    def get_current_task_name(self) -> str:
        """Return the name of the task currently executing.

        This method can only be called in a Snowflake task.

        Example:
            >>> def task_handler(session: Session) -> None:
            >>>     from snowflake.core.task.context import TaskContext
            >>>     context = TaskContext(session)
            >>>     task_name = context.get_current_task_name()
        """
        if not self._current_task_name:
            self._current_task_name = str(self._session.call("system$current_user_task_name"))
        return self._current_task_name

    def get_current_task_short_name(self) -> str:
        """Return the task name under the context of a DAG.

        If a :class:`snowflake.core.task.dagv1.DAGTask` is created under a :class:`snowflake.core.task.dagv1.DAG`
        object, the real task name is in format ``{dag name}${task name}``. This API returns the ``task name`` part.
        :meth:`get_current_task_name` returns the real task.
        """
        root_prefix = self._get_root_prefix()
        name = self.get_current_task_name()
        name = name[name.rfind(".") + 1 :]
        if root_prefix and name.startswith(root_prefix):
            return name[len(root_prefix) :]
        return name

    def get_task_graph_config(self) -> Optional[dict[str, Any]]:
        """Return the graph config of the task."""
        config = self._session.call("SYSTEM$GET_TASK_GRAPH_CONFIG")
        if config:
            return json.loads(config)  # type: ignore
        return None

    def get_task_graph_config_property(self, property_name: str) -> Optional[Any]:
        """Return the graph config of the task."""
        dag_config = self.get_task_graph_config()
        return dag_config.get(property_name) if dag_config else None

    def get_current_root_task_name(self) -> str:
        """Return the current task's root task name."""
        if not self._current_root_name:
            self._current_root_name = self.get_runtime_info("CURRENT_ROOT_TASK_NAME")  # type: ignore
        return self._current_root_name  # type: ignore

    def get_current_root_task_uuid(self) -> str:
        """Return the current task's root task UUID."""
        return self.get_runtime_info("CURRENT_ROOT_TASK_UUID")  # type: ignore

    def get_current_task_graph_original_schedule(self) -> str:
        """Return the current task's original schedule for this run or its root task's schedule."""
        return self.get_runtime_info("CURRENT_TASK_GRAPH_ORIGINAL_SCHEDULED_TIMESTAMP")  # type: ignore

    def get_current_task_graph_run_group_id(self) -> str:
        """Return the current task's run group."""
        return self.get_runtime_info("CURRENT_TASK_GRAPH_RUN_GROUP_ID")  # type: ignore

    def get_last_successful_task_graph_original_schedule(self) -> Optional[str]:
        """Return the last successful task run schedule."""
        return self.get_runtime_info("LAST_SUCCESSFUL_TASK_GRAPH_ORIGINAL_SCHEDULED_TIMESTAMP")  # type: ignore

    def get_last_successful_task_graph_run_group_id(self) -> Optional[str]:
        """Return the last successful task run group id."""
        return self.get_runtime_info("LAST_SUCCESSFUL_TASK_GRAPH_RUN_GROUP_ID")  # type: ignore

    def get_runtime_info(self, property_name: str) -> Optional[Union[str, datetime]]:
        """Return the runtime information of the current task.

        You usually don't need to call this function. Call the other ``get_*`` functions instead.
        """
        if not property_name:
            raise ValueError("`property_name` must be an non-empty str.")
        property_name_upper = property_name.upper()

        try:
            if property_name in (
                "CURRENT_TASK_GRAPH_ORIGINAL_SCHEDULED_TIMESTAMP",
                "LAST_SUCCESSFUL_TASK_GRAPH_ORIGINAL_SCHEDULED_TIMESTAMP",
            ):
                result = self._session.sql(
                    f"select to_timestamp(system$task_runtime_info('{property_name_upper}'))"
                ).collect()[0][0]
            else:
                result = self._session.sql(
                    f"select to_char(system$task_runtime_info('{property_name_upper}'))"
                ).collect()[0][0]
            return result
        except Exception as sse:
            if "NULL result in a non-nullable column" in str(sse):
                return None
            raise sse


__all__ = ["TaskContext"]
